"""
Adapter for the Twisted network programming framework.
"""

try:
    import twisted
except ImportError:
    twisted = None

if not twisted:
    raise ImportError("Twisted is required for this module to work: "
                      "https://twistedmatrix.com/")

from twisted.internet.protocol import Factory

from ..twistedsupport import TwistedAdapter


if __name__ == "__main__":
    import logging

    from twisted.internet import reactor
    from twisted.internet.endpoints import TCP4ClientEndpoint
    from twisted.internet.error import ReactorNotRunning

    from . import examplecli

    parser = examplecli.create_parser(
        description="Twisted-based AMI client example")

    options, args = examplecli.parse_args(parser)

    log = logging.getLogger(__name__)

    # connectProtocol() appeared in 13.1, but we want to support at least 11.0+.
    # (endpoints appeared in 10.1).

    class CLIFactory(Factory):
        def buildProtocol(self, addr):
            proto = examplecli.CLIProtocol(reactor, options)
            return TwistedAdapter(proto)

    endpoint = TCP4ClientEndpoint(reactor, options.host, options.port)
    d = endpoint.connect(CLIFactory())

    def connection_failed(failure):
        log.error("Connection to %r failed: %s",
                  (options.host, options.port), failure)
        reactor.stop()

    # Hack reactor.stop() to mute errors on double stop
    _old_reactor_stop = reactor.stop
    def _reactor_stop():
        try:
            _old_reactor_stop()
        except ReactorNotRunning:
            pass
    reactor.stop = _reactor_stop

    d.addErrback(connection_failed)
    reactor.run()



"""
FastAGI adapter for the Tornado network programming framework.
"""

try:
    import tornado
except ImportError:
    tornado = None

if not tornado:
    raise ImportError("tornado is required for this module to work: "
                      "http://www.tornadoweb.org/")

from tornado import netutil
try:
    from tornado.tcpserver import TCPServer
except ImportError:
    from tornado.netutil import TCPServer

from ..tornadosupport import TornadoAdapter


class FastAGIServer(TCPServer):
    """
    A TCPServer subclass that will delegate all incoming connections
    to a FastAGIExecutor instance.
    """

    def __init__(self, executor, *args, **kwargs):
        TCPServer.__init__(self, *args, **kwargs)
        self.executor = executor

    def handle_stream(self, stream, remote_addr):
        proto = self.executor.make_protocol()
        adapter = TornadoAdapter(proto)
        adapter.bind_stream(stream)


if __name__ == "__main__":
    from tornado.ioloop import IOLoop

    from .fastagi import FastAGIProtocol, FastAGIExecutor, TCP_PORT
    from .session import AGISession
    from . import examplecli

    parser = examplecli.create_parser(
        description="Tornado-based FastAGI server example")
    parser.add_argument('-p', '--port', type=int, default=TCP_PORT,
                        help='port to listen on')
    parser.add_argument('-L', '--listen', default='127.0.0.1',
                        help='address to listen on')

    options, args = examplecli.parse_args(parser)

    loop = IOLoop.instance()

    class CLIProtocol(examplecli.CLIProtocol, FastAGIProtocol):
        pass

    executor = FastAGIExecutor(CLIProtocol)
    server = FastAGIServer(executor, io_loop=loop)
    server.listen(args.port, args.listen)

    try:
        loop.start()
    except KeyboardInterrupt:
        pass




import unittest

from obelus.casedict import CaseDict
from obelus.test import main


class CaseDictTest(unittest.TestCase):

    def check_underlying_dict(self, d, expected):
        """
        Check for implementation details.
        """
        self.assertEqual(set(d._data), set(expected))
        self.assertEqual({k: d[k] for k in d._data}, expected)
        self.assertEqual(len(d._data), len(expected))

    # NOTE: we only test the operations which are not inherited from
    # MutableMapping.

    def test_init(self):
        d = CaseDict()
        self.check_underlying_dict(d, {})
        d = CaseDict({'Foo': 5, 'baR': 6})
        self.check_underlying_dict(d, {'foo': 5, 'bar': 6})
        d = CaseDict(fOO=5, Bar=6)
        self.check_underlying_dict(d, {'foo': 5, 'bar': 6})
        d = CaseDict({'FOO': 5}, Bar=6)
        self.check_underlying_dict(d, {'foo': 5, 'bar': 6})

    def test_setitem_getitem(self):
        d = CaseDict()
        with self.assertRaises(KeyError):
            d['foo']
        d['Foo'] = 5
        self.assertEqual(d['foo'], 5)
        self.assertEqual(d['Foo'], 5)
        self.assertEqual(d['FOo'], 5)
        with self.assertRaises(KeyError):
            d['bar']
        self.check_underlying_dict(d, {'foo': 5})
        d['BAR'] = 6
        self.assertEqual(d['Bar'], 6)
        self.check_underlying_dict(d, {'foo': 5, 'bar': 6})
        # Overwriting
        d['foO'] = 7
        self.assertEqual(d['foo'], 7)
        self.assertEqual(d['Foo'], 7)
        self.assertEqual(d['FOo'], 7)
        self.check_underlying_dict(d, {'foo': 7, 'bar': 6})

    def test_delitem(self):
        d = CaseDict(Foo=5)
        d['baR'] = 3
        del d['fOO']
        self.check_underlying_dict(d, {'bar': 3})
        with self.assertRaises(KeyError):
            del d['Foo']
        with self.assertRaises(KeyError):
            del d['foo']

    def test_get(self):
        d = CaseDict()
        default = object()
        self.assertIs(d.get('foo'), None)
        self.assertIs(d.get('foo', default), default)
        d['Foo'] = 5
        self.assertEqual(d.get('foo'), 5)
        self.assertEqual(d.get('FOO'), 5)
        self.assertIs(d.get('bar'), None)
        self.check_underlying_dict(d, {'foo': 5})

    def test_pop(self):
        d = CaseDict()
        default = object()
        with self.assertRaises(KeyError):
            d.pop('foo')
        self.assertIs(d.pop('foo', default), default)
        d['Foo'] = 5
        self.assertIn('foo', d)
        self.assertEqual(d.pop('foo'), 5)
        self.assertNotIn('foo', d)
        self.check_underlying_dict(d, {})
        d['Foo'] = 5
        self.assertIn('Foo', d)
        self.assertEqual(d.pop('FOO'), 5)
        self.assertNotIn('foo', d)
        self.check_underlying_dict(d, {})
        with self.assertRaises(KeyError):
            d.pop('foo')

    def test_clear(self):
        d = CaseDict()
        d.clear()
        self.check_underlying_dict(d, {})
        d['Foo'] = 5
        d['baR'] = 3
        self.check_underlying_dict(d, {'foo': 5, 'bar': 3})
        d.clear()
        self.check_underlying_dict(d, {})

    def test_contains(self):
        d = CaseDict()
        self.assertIs(False, 'foo' in d)
        d['Foo'] = 5
        self.assertIs(True, 'Foo' in d)
        self.assertIs(True, 'foo' in d)
        self.assertIs(True, 'FOO' in d)
        self.assertIs(False, 'bar' in d)

    def test_len(self):
        d = CaseDict()
        self.assertEqual(len(d), 0)
        d['Foo'] = 5
        self.assertEqual(len(d), 1)
        d['BAR'] = 6
        self.assertEqual(len(d), 2)
        d['foo'] = 7
        self.assertEqual(len(d), 2)
        d['baR'] = 3
        self.assertEqual(len(d), 2)
        del d['Bar']
        self.assertEqual(len(d), 1)

    def test_iter(self):
        d = CaseDict()
        it = iter(d)
        with self.assertRaises(StopIteration):
            next(it)
        d['Foo'] = 5
        d['BAR'] = 6
        yielded = []
        for x in d:
            yielded.append(x)
        self.assertEqual(set(yielded), {'Foo', 'BAR'})

    def test_repr(self):
        d = CaseDict()
        self.assertEqual(repr(d), "CaseDict()")
        d['Foo'] = 5
        self.assertEqual(repr(d), "CaseDict({'Foo': 5})")
        d['Bar'] = 6
        self.assertIn(repr(d), ("CaseDict({'Foo': 5, 'Bar': 6})",
                                "CaseDict({'Bar': 6, 'Foo': 5})"))


if __name__ == "__main__":
    main()





class TornadoAdapter(object):
    """
    Tornado adapter for Obelus protocols (e.g. AMIProtocol, AGIProtocol).

    Pass a *protocol* instance to create the adapter, then call
    :meth:`bind_stream` when you need to wire the protocol to a Tornado
    :class:`~tornado.iostream.IOStream` instance.
    """

    stream = None

    def __init__(self, protocol):
        self.protocol = protocol

    def bind_stream(self, stream):
        """
        Bind the protocol to the given IOStream.  The stream should
        be already connected, as the protocol's connection_made() will
        be called immediately.
        """
        self.stream = stream
        self.stream.read_until_close(self._final_cb, self._streaming_cb)
        self.stream.set_close_callback(self._close_cb)
        self.protocol.connection_made(self)

    def _streaming_cb(self, data):
        if data:
            self.protocol.data_received(data)

    def _final_cb(self, data):
        # This is called when reading is finished, either because
        # of a regular EOF or because of an error.
        self._streaming_cb(data)
        self._close_cb()

    def _close_cb(self, *args):
        if self.stream is not None:
            stream = self.stream
            self.stream = None
            self.protocol.connection_lost(stream.error)

    def write(self, data):
        if self.stream is None:
            raise ValueError("write() on a non-connected protocol")
        self.stream.write(data)

    def close(self):
        if self.stream is not None:
            self.stream.close()






from collections import OrderedDict
import textwrap
import unittest

from mock import Mock, ANY

from obelus.ami.protocol import (
    BaseAMIProtocol, AMIProtocol, Event, Response, EventList, ActionError)
from obelus.common import Handler
from obelus.test import main


def literal_message(text):
    return textwrap.dedent(text.rstrip() + '\n\n').encode('utf-8')

EVENT_HANGUP = literal_message("""\
    Event: Hangup
    Privilege: call,all
    Channel: SIP/0004F2060EB4-00000000
    Uniqueid: 1283174108.0
    """)

CORE_SETTINGS_RESPONSE = literal_message("""\
    Response: Success
    AMIversion: 1.1
    AsteriskVersion: 1.8.13.0~dfsg-1
    """)

ERROR_RESPONSE = literal_message("""\
    Response: Error
    ActionID: 4444
    Message: Invalid/unknown command: xyzzy. (blabla)
    """)

LOGOFF_RESPONSE = literal_message("""\
    Response: Goodbye
    Message: Thanks for all the fish.
    """)

CORE_SHOW_VERSION_RESPONSE = literal_message("""\
    Response: Follows
    Privilege: Command
    ActionID: DEF.768
    Asterisk 1.8.13.0~dfsg-1
    --END COMMAND--
    """)

OTHER_COMMAND_RESPONSE = literal_message("""\
    Response: Follows
    Privilege: Command
    ActionID: 1234
    foo
    bar--END COMMAND--
    """)

DIALPLAN_START_RESPONSE = literal_message("""\
    Response: Success
    ActionID: 123.567
    EventList: start
    Message: DialPlan list will follow
    """)

DIALPLAN_RESPONSE_EVENTS = literal_message("""\
    Event: ListDialplan
    ActionID: 123.567
    Context: inbound-call
    Registrar: pbx_config

    Event: ListDialplan
    ActionID: 123.567
    Context: default
    IncludeContext: outgoing-call-leg1
    Registrar: pbx_config
    """)

DIALPLAN_EVENTS_END = literal_message("""\
    Event: ShowDialPlanComplete
    EventList: Complete
    ListItems: 52
    ListExtensions: 19
    ListPriorities: 51
    ListContexts: 19
    ActionID: 123.567
    """)


class ProtocolTestBase(object):

    greeting_line = b"Asterisk Call Manager/1.4\r\n"

    def setUp(self):
        p = self.proto = self.protocol_factory()
        p.greeting_received = Mock()

    def ready_proto(self):
        p = self.proto
        p.data_received(self.greeting_line)
        return p

    def feed(self, message):
        lines = message.splitlines(True)
        for line in lines:
            self.proto.line_received(line)

    def assert_called_once_with_exc(self, callback, exc_class):
        callback.assert_called_once_with(ANY)
        (exc,), _ = callback.call_args
        return exc


class BaseAMIProtocolTest(ProtocolTestBase, unittest.TestCase):

    protocol_factory = BaseAMIProtocol

    def test_greeting_line(self):
        p = self.proto
        self.assertEqual(p._state, 'init')
        self.assertEqual(p.greeting_received.call_count, 0)
        p.data_received(self.greeting_line)
        p.greeting_received.assert_called_once_with(
            "Asterisk Call Manager", "1.4")
        self.assertEqual(p._state, 'idle')

    def test_event_received(self):
        p = self.ready_proto()
        p.event_received = Mock()
        expected = Event('Hangup', {
            'Privilege': 'call,all',
            'Channel': 'SIP/0004F2060EB4-00000000',
            'Uniqueid': '1283174108.0',
            })
        self.feed(EVENT_HANGUP)
        p.event_received.assert_called_once_with(expected)
        (evt,), _ = p.event_received.call_args
        # Headers are case-insensitive
        self.assertEqual(evt.headers['privilege'], 'call,all')
        self.assertEqual(evt.headers['Privilege'], 'call,all')
        self.assertEqual(p._state, 'idle')
        p.event_received.reset_mock()
        self.feed(EVENT_HANGUP)
        p.event_received.assert_called_once_with(expected)
        self.assertEqual(p._state, 'idle')

    def test_response_received(self):
        p = self.ready_proto()
        p.response_received = Mock()
        core_settings = Response('success', {
            'AMIversion': '1.1',
            'AsteriskVersion': '1.8.13.0~dfsg-1',
            }, [])
        error = Response('error', {
            'ActionID': '4444',
            'Message': 'Invalid/unknown command: xyzzy. (blabla)',
            }, [])
        goodbye = Response('goodbye', {
            'Message': 'Thanks for all the fish.',
            }, [])
        self.feed(CORE_SETTINGS_RESPONSE)
        p.response_received.assert_called_once_with(core_settings)
        (resp,), _ = p.response_received.call_args
        # Headers are case-insensitive
        self.assertEqual(resp.headers['amiversion'], '1.1')
        self.assertEqual(resp.headers['AmiVersion'], '1.1')
        p.response_received.reset_mock()
        self.feed(ERROR_RESPONSE)
        p.response_received.assert_called_once_with(error)
        p.response_received.reset_mock()
        self.feed(CORE_SETTINGS_RESPONSE)
        p.response_received.assert_called_once_with(core_settings)
        p.response_received.reset_mock()
        self.feed(LOGOFF_RESPONSE)
        p.response_received.assert_called_once_with(goodbye)
        self.assertEqual(p._state, 'idle')

    def test_command_response_received(self):
        p = self.ready_proto()
        p.response_received = Mock()
        expected = Response('follows', {
            'Privilege': 'Command',
            'ActionID': 'DEF.768',
            }, ['Asterisk 1.8.13.0~dfsg-1'])
        self.feed(CORE_SHOW_VERSION_RESPONSE)
        p.response_received.assert_called_once_with(expected)
        self.assertEqual(p._state, 'idle')

    def test_command_response_received_no_eol(self):
        # Sometimes the command payload is concatenated to the
        # '--END COMMAND--' marker without a EOL separator.
        p = self.ready_proto()
        p.response_received = Mock()
        expected = Response('follows', {
            'Privilege': 'Command',
            'ActionID': '1234',
            }, ['foo', 'bar'])
        self.feed(OTHER_COMMAND_RESPONSE)
        p.response_received.assert_called_once_with(expected)
        self.assertEqual(p._state, 'idle')

    def test_serialize_message(self):
        p = self.ready_proto()
        expected = b"foo: bar\r\n\r\n"
        self.assertEqual(expected, p.serialize_message({'foo': 'bar'}))
        expected = (
            b"foo: bar\r\n"
            b"foo: zyxxy\r\n"
            b"\r\n")
        self.assertEqual(expected,
                         p.serialize_message({'foo': ['bar', 'zyxxy']}))
        expected = (
            b"foo: bar\r\n"
            b"quux: zyxxy\r\n"
            b"\r\n")
        headers = OrderedDict([('foo', 'bar'), ('quux', 'zyxxy')])
        self.assertEqual(expected, p.serialize_message(headers))

    def test_serialize_message_type_error(self):
        p = self.ready_proto()
        with self.assertRaises(TypeError):
            p.serialize_message({'foo': 1})
        if str is not bytes:
            with self.assertRaises(TypeError):
                p.serialize_message({b'foo': b'bar'})


class AMIProtocolTest(ProtocolTestBase, unittest.TestCase):

    protocol_factory = AMIProtocol

    def test_event_received(self):
        p = self.ready_proto()
        p.unhandled_event_received = Mock()
        self.feed(EVENT_HANGUP)
        p.unhandled_event_received.assert_called_once_with(
            Event('Hangup', {
                'Privilege': 'call,all',
                'Channel': 'SIP/0004F2060EB4-00000000',
                'Uniqueid': '1283174108.0',
                }))

    def test_event_handler(self):
        p = self.ready_proto()
        cb_hangup, cb_foobar = Mock(), Mock()
        p.register_event_handler('Hangup', cb_hangup)
        p.register_event_handler('Foobar', cb_foobar)
        self.feed(EVENT_HANGUP)
        cb_hangup.assert_called_once_with(
            Event('Hangup', {
                'Privilege': 'call,all',
                'Channel': 'SIP/0004F2060EB4-00000000',
                'Uniqueid': '1283174108.0',
                }))

    def test_send_action(self):
        p = self.ready_proto()
        p.write = Mock()
        a = p.send_action('Hello', OrderedDict({'foo': 'bar'}))
        self.assertIsInstance(a, Handler)
        p.write.assert_called_once_with(
            b"foo: bar\r\n"
            b"Action: Hello\r\n"
            b"ActionID: 1\r\n"
            b"\r\n")
        self.assertEqual(a._action_id, '1')
        a = p.send_action('Hi', {'Channel': 'SIP/foo'})
        self.assertIsInstance(a, Handler)
        (data,), _ = p.write.call_args
        lines = data.splitlines()
        self.assertIn(b"ActionID: 2", lines)
        self.assertIn(b"Action: Hi", lines)
        self.assertIn(b"Channel: SIP/foo", lines)
        self.assertEqual(a._action_id, '2')
        self.assertEqual(set(p._actions), {'1', '2'})

    def test_send_action_variables(self):
        p = self.ready_proto()
        p.write = Mock()
        a = p.send_action('Hi', OrderedDict(), {})
        (data,), _ = p.write.call_args
        self.assertEqual(data, b"Action: Hi\r\nActionID: 1\r\n\r\n")
        a = p.send_action('Hi', {'x': 'y'}, {'foo': '1', 'bar': '2'})
        (data,), _ = p.write.call_args
        lines = data.splitlines(True)
        self.assertEqual(set(lines), {b"x: y\r\n",
                                      b"Variable: foo=1\r\n",
                                      b"Variable: bar=2\r\n",
                                      b"Action: Hi\r\n",
                                      b"ActionID: 2\r\n",
                                      b"\r\n"
                                      })

    def test_send_action_override_action_id(self):
        p = self.ready_proto()
        p.write = Mock()
        a = p.send_action('Hi', {'ActionID': 'ABCD'})
        (data,), _ = p.write.call_args
        lines = data.splitlines(True)
        self.assertEqual(set(lines), {b"Action: Hi\r\n",
                                      b"ActionID: ABCD\r\n",
                                      b"\r\n"
                                      })
        self.assertEqual(a._action_id, 'ABCD')
        self.assertEqual(set(p._actions), {'ABCD'})

    def test_send_action_response_success(self):
        p = self.ready_proto()
        p.write = Mock()
        a = p.send_action('CoreSettings', {'ActionID': '1234'})
        a.on_result = Mock()
        a.on_exception = Mock()
        resp = literal_message("""\
            Response: Success
            ActionID: 1234
            AsteriskVersion: 1.8.13
            """)
        p.data_received(resp)
        a.on_result.assert_called_once_with(
            Response('success', {'ActionID': '1234',
                                 'AsteriskVersion': '1.8.13'}, []))
        self.assertEqual(a.on_exception.call_count, 0)
        # The handler has been unregistered: no further callbacks
        p.data_received(resp)
        self.assertEqual(a.on_result.call_count, 1)
        self.assertEqual(a.on_exception.call_count, 0)

    def test_send_action_response_follows(self):
        p = self.ready_proto()
        p.write = Mock()
        a = p.send_action('Command',
                          {'Command': 'core show version',
                           'ActionID': 'DEF.768'})
        a.on_result = Mock()
        a.on_exception = Mock()
        p.data_received(CORE_SHOW_VERSION_RESPONSE)
        a.on_result.assert_called_once_with(
            Response('follows', {'ActionID': 'DEF.768',
                                 'Privilege': 'Command'},
                                 ['Asterisk 1.8.13.0~dfsg-1']))
        self.assertEqual(a.on_exception.call_count, 0)

    def test_send_action_response_goodbye(self):
        p = self.ready_proto()
        p.write = Mock()
        a = p.send_action('Logoff',
                          {'ActionID': '1234'})
        a.on_result = Mock()
        a.on_exception = Mock()
        resp = literal_message("""\
            Response: Goodbye
            ActionID: 1234
            Message: Thanks for all the fish.
            """)
        p.data_received(resp)
        a.on_result.assert_called_once_with(
            Response('goodbye', {'ActionID': '1234',
                                 'Message': 'Thanks for all the fish.'},
                                []))
        self.assertEqual(a.on_exception.call_count, 0)

    def test_send_action_response_error(self):
        p = self.ready_proto()
        p.write = Mock()
        a = p.send_action('xyzzy',
                          {'ActionID': '4444'})
        a.on_result = Mock()
        a.on_exception = Mock()
        p.data_received(ERROR_RESPONSE)
        self.assertEqual(a.on_result.call_count, 0)
        exc = self.assert_called_once_with_exc(a.on_exception, ActionError)
        self.assertEqual(str(exc), "Invalid/unknown command: xyzzy. (blabla)")

    def test_send_action_event_list(self):
        p = self.ready_proto()
        p.write = Mock()
        a = p.send_action('ShowDialPlan',
                          {'ActionID': '123.567'})
        a.on_result = Mock()
        a.on_exception = Mock()
        p.data_received(DIALPLAN_START_RESPONSE)
        self.assertEqual(a.on_result.call_count, 0)
        self.assertEqual(a.on_exception.call_count, 0)
        p.data_received(DIALPLAN_RESPONSE_EVENTS)
        self.assertEqual(a.on_result.call_count, 0)
        self.assertEqual(a.on_exception.call_count, 0)
        p.data_received(DIALPLAN_EVENTS_END)
        a.on_result.assert_called_once_with(ANY)
        self.assertEqual(a.on_exception.call_count, 0)
        (evlist,), _ = a.on_result.call_args
        self.assertIsInstance(evlist, EventList)
        self.assertEqual(len(evlist.events), 2)
        self.assertEqual(evlist.events[0], Event(
            name='ListDialplan',
            headers={
                'ActionID': '123.567',
                'Context': 'inbound-call',
                'Registrar': 'pbx_config',
                }))
        self.assertEqual(evlist.events[1], Event(
            name='ListDialplan',
            headers={
                'ActionID': '123.567',
                'Context': 'default',
                'IncludeContext': 'outgoing-call-leg1',
                'Registrar': 'pbx_config',
                }))
        # End headers are merged with start headers
        self.assertEqual(evlist.headers, {
            'ActionID': '123.567',
            'EventList': 'Complete',
            'ListItems': '52',
            'ListExtensions': '19',
            'ListPriorities': '51',
            'ListContexts': '19',
            'ActionID': '123.567',
            'Message': 'DialPlan list will follow',
            })



if __name__ == "__main__":
    main()



"""
Event-driven, framework-agnostic implementation of the AMI protocol.
"""

import collections
import logging

from ..casedict import CaseDict
from ..common import Handler, LineReceiver


_BaseResponse = collections.namedtuple('_BaseResponse',
                                       ('type', 'headers', 'payload'))

class Response(_BaseResponse):
    __slots__ = ()


_BaseEvent = collections.namedtuple('_BaseEvent',
                                    ('name', 'headers'))

class Event(_BaseEvent):
    __slots__ = ()


_BaseEventList = collections.namedtuple('_BaseEventList',
                                        ('headers', 'events'))

class EventList(_BaseEventList):
    __slots__ = ()


class ActionError(RuntimeError):
    """
    An error response was received following an action.
    """


class BaseAMIProtocol(LineReceiver):
    """
    Implementation of the AMI protocol syntax.
    """

    # XXX The AMI charset isn't really defined, it seems Asterisk
    # will just pass bytestrings around without caring.  We use
    # utf-8 as a reasonable default for common setups.
    encoding = 'utf-8'
    eol = '\r\n'

    logger = logging.getLogger(__name__)
    transport = None
    # If set to True, all incoming messages will be logged in debug level
    trace_messages = False

    _response_types = {'success', 'follows', 'error', 'goodbye'}
    _headers_for_response_follows = {'Privilege', 'ActionID'}
    _response_follows_end = '--END COMMAND--'

    def __init__(self):
        super(BaseAMIProtocol, self).__init__()
        self.reset()

    def reset(self):
        self._state = 'init'
        self._event_handlers = {}

    def _split_key_value(self, line):
        key, sep, value = line.rstrip().partition(':')
        if not sep:
            raise ValueError("Expected a key/value pair, got %r"
                             % (line,))
        return key, value.lstrip()

    def connection_made(self, transport):
        self.transport = transport

    def connection_lost(self, exc):
        pass

    def write(self, data):
        self.transport.write(data)

    def greeting_received(self, api_name, api_version):
        """
        Called when the AMI's initial greeting line is received.
        Typical values for *api_name* and *api_version* are
        "Asterisk Call Manager" and "1.1", respectively.

        Override this method to do something with said information.
        """

    def line_received(self, line):
        """
        Processing an incoming *line* of AMI data.
        """
        line = line.rstrip(b'\r\n')
        if not isinstance(line, str):
            # Python 3 only
            line = line.decode(self.encoding)
        if self._state == 'init':
            # Parse greeting line
            name, version = line.strip().split("/")
            self._state = 'idle'
            self.greeting_received(name, version)
        elif self._state == 'idle':
            # Start of event or response
            if not line:
                return
            key, value = self._split_key_value(line)
            if key == 'Response':
                self._state = 'in-response'
                self._headers = CaseDict()
                self._payload = []
                self._resp_type = value.lower()
                if self._resp_type not in self._response_types:
                    raise ValueError("Invalid response type %r"
                                     % (self.resp_type))
            elif key == 'Event':
                self._state = 'in-event'
                self._headers = CaseDict()
                self._event_type = value
            else:
                raise ValueError("Unexpected first message line %r" % (line,))
        elif self._state == 'in-response':
            if not line:
                # Response ends with an empty line
                self._state = 'idle'
                self._response_complete()
            else:
                # Expect a "Key: value" line
                key, value = self._split_key_value(line)
                self._headers[key] = value
                if (self._resp_type == 'follows' and
                    set(self._headers) >= self._headers_for_response_follows):
                    # The payload of a "command" response comes after the
                    # "Response", "ActionID" and "Privilege" headers.
                    assert not self._payload
                    self._state = 'in-response-follows'
        elif self._state == 'in-event':
            if not line:
                # Event ends with an empty line
                self._state = 'idle'
                self._event_complete()
            else:
                # Expect a "Key: value" line
                key, value = self._split_key_value(line)
                self._headers[key] = value
        elif self._state == 'in-response-follows':
            # Inside a command payload: accumulate until we encounter
            # the '--END COMMAND--' marker.
            if line.endswith(self._response_follows_end):
                line = line[:-len(self._response_follows_end)]
                if line:
                    self._payload.append(line)
                self._state = 'idle'
                self._response_complete()
            else:
                self._payload.append(line)
        else:
            # Shouldn't come here
            assert 0

    def _response_complete(self):
        resp = Response(self._resp_type, self._headers, self._payload)
        self.logger.debug("Received response: %r", resp)
        self.response_received(resp)

    def _event_complete(self):
        event = Event(self._event_type, self._headers)
        if self.trace_messages:
            self.logger.debug("Received event: %r", event)
        self.event_received(event)

    def response_received(self, resp):
        """
        Called when a response is received.
        """

    def event_received(self, event):
        """
        Called when an event is received.
        """

    def serialize_message(self, headers):
        """
        Serialize a message comprised of the given *headers*.
        Each header value can be either a str or a list of str objects.
        """
        lines = []
        for key, values in headers.items():
            if not isinstance(values, list):
                values = [values]
            for value in values:
                lines.append(": ".join([key, value]))
        lines.extend(["", ""])
        data = self.eol.join(lines)
        if not isinstance(data, bytes):
            data = data.encode(self.encoding)
        return data


class AMIProtocol(BaseAMIProtocol):
    """
    Higher-level AMI protocol implementation, including response and event
    matching.
    """

    def reset(self):
        super(AMIProtocol, self).reset()
        self._action_id = 1
        self._actions = {}
        self._event_lists = {}

    def _next_action_id(self):
        a = self._action_id
        self._action_id = a + 1
        return str(a)

    def send_action(self, name, headers, variables=()):
        """
        Send the AMI action with the given *name* (a str object)
        and *headers* (a dict mapping names onto values).
        Return a Handler which will be fired when the AMI returns a
        response for the action.
        """
        if variables:
            vars_list = headers.setdefault('Variable', [])
            for key, value in variables.items():
                vars_list.append('='.join([key, value]))
        headers['Action'] = name
        try:
            action_id = headers['ActionID']
        except KeyError:
            action_id = headers['ActionID'] = self._next_action_id()
        data = self.serialize_message(headers)
        self.logger.debug("Sending action: %r", data)
        self.write(data)
        handler = Handler()
        handler._action_id = action_id
        self._actions[action_id] = handler
        return handler

    def register_event_handler(self, name, func):
        """
        Register a callable event handler *func* for the event *name*.
        """
        if not isinstance(name, str):
            raise TypeError("Event name should be str, not %r" % (name.__class__))
        if name in self._event_handlers:
            raise KeyError("Handler already registered for %r" % (name,))
        self._event_handlers[name] = func

    def unregister_event_handler(self, name):
        """
        Unregister the handler for event *name*.
        """
        del self._event_handlers[name]

    def _handle_event_list_start(self, resp, action_id, handler):
        if action_id in self._event_lists:
            self.logger.error("Received new EventList for "
                              "ungoing event list %r" % (action_id,))
            return
        self._event_lists[action_id] = EventList(resp.headers, [])

    def event_received(self, event):
        action_id = event.headers.get('ActionID')
        action_handler = action_id and self._actions.get(action_id)
        event_list = action_handler and self._event_lists.get(action_id)
        if event_list is not None:
            # Handle potential event list item
            event_type = event.headers.get('EventList', '').lower()
            if event_type == 'complete':
                del self._event_lists[action_id]
                del self._actions[action_id]
                # We merge the end event's headers into the EventList's,
                # since they can carry useful information.
                event_list.headers.update(event.headers)
                action_handler.set_result(event_list)
                return
            elif event_type:
                self.logger.warn("Invalid EventList header in event: %r"
                                 % (event_type,))
            event_list.events.append(event)
            return
        self._dispatch_event(event)

    def _dispatch_event(self, event):
        try:
            handler = self._event_handlers[event.name]
        except KeyError:
            self.unhandled_event_received(event)
        else:
            handler(event)

    def unhandled_event_received(self, event):
        """
        Called when an *event* is received for which no handler has
        been registered.
        """
        self.logger.debug("Unhandled event type %r", event.name)

    def response_received(self, resp):
        action_id = resp.headers['ActionID']
        try:
            handler = self._actions[action_id]
        except KeyError:
            self.logger.info("Unknown or stale response received "
                             "with action ID %r" % (action_id,))
            return
        if resp.type == 'error':
            del self._actions[action_id]
            exc = ActionError(resp.headers.get('Message', ''))
            handler.set_exception(exc)
        elif resp.type in ('success', 'goodbye', 'follows'):
            event_type = resp.headers.get('EventList', '').lower()
            if event_type == 'start':
                self._handle_event_list_start(resp, action_id, handler)
                return
            elif event_type:
                self.logger.warn("Invalid EventList header in response: %r"
                                 % (event_type,))
            del self._actions[action_id]
            handler.set_result(resp)
        else:
            # Can't come here
            assert 0




"""
Adapter for the Tulip network programming framework.
"""

try:
    import tulip
except ImportError:
    tulip = None

if not tulip:
    raise ImportError("tulip is required for this module to work: "
                      "http://code.google.com/p/tulip/")


if __name__ == "__main__":
    import logging
    import signal

    from . import examplecli

    parser = examplecli.create_parser(
        description="Tulip-based AMI client example")

    options, args = examplecli.parse_args(parser)

    # Tulip's logger is very chatty, dampen it
    logging.getLogger('tulip').setLevel('WARNING')
    log = logging.getLogger(__name__)

    loop = tulip.get_event_loop()
    proto = examplecli.CLIProtocol(loop, options)
    fut = tulip.async(loop.create_connection(lambda: proto,
                                             options.host, options.port))
    def cb(fut):
        exc = fut.exception()
        if exc is not None:
            log.error("Connection to %r failed: %s",
                      (options.host, options.port), exc)
            loop.stop()
    fut.add_done_callback(cb)

    loop.add_signal_handler(signal.SIGINT, loop.stop)
    loop.run_forever()



import weakref

from .protocol import AGIProtocol, ProtocolAGIChannel


TCP_PORT = 4573


class FastAGIExecutor(object):

    def __init__(self, protocol_factory):
        self._conns = weakref.WeakSet()
        self.protocol_factory = protocol_factory

    def make_protocol(self):
        proto = self.protocol_factory(ProtocolAGIChannel())
        self._conns.add(proto)
        return proto

    def close(self):
        for proto in list(self._conns):
            proto.transport.close()
            proto.connection_lost(None)


class FastAGIProtocol(AGIProtocol):

    def connection_made(self, transport):
        super(FastAGIProtocol, self).connection_made(transport)
        self.bind_session()

    def connection_lost(self, exc):
        super(FastAGIProtocol, self).connection_lost(exc)
        self.unbind_session()



"""
Event-driven implementation of the AGI protocol.
"""

import collections
import logging
import re

from ..common import Handler, LineReceiver


class AGIError(RuntimeError):
    """
    An error returned by Asterisk through the AGI channel.
    """

class AGIUnknownCommand(AGIError):
    """
    Unknown AGI command.
    """

class AGIForbiddenCommand(AGIError):
    """
    AGI command disallowed on a dead channel.
    """

class AGISyntaxError(AGIError):
    """
    Invalid syntax or arguments for an AGI command.
    """

class AGICommandFailure(AGIError):
    """
    Negative result code returned by a command.
    """

_agi_errors = {
    510: AGIUnknownCommand,
    511: AGIForbiddenCommand,
    520: AGISyntaxError,
}


_BaseResponse = collections.namedtuple('_BaseResponse',
                                       ('result', 'variables', 'data'))

class Response(_BaseResponse):
    __slots__ = ()


class AGIChannel(object):

    def send_command_line(self, line):
        raise NotImplementedError


class ProtocolAGIChannel(AGIChannel):
    """
    An AGI channel which simply re-uses its protocol for communications.
    Meant to be used for the "normal" forms of AGI (script AGI, Fast AGI).
    """

    def send_command_line(self, line):
        handler = Handler()
        self.write(line)
        # Expect responses in sequential order.
        self.proto._push_command(handler)
        return handler

    def write(self, data):
        self.proto.write(data)


class AGIProtocol(LineReceiver):

    # XXX The AGI charset isn't really defined, it seems Asterisk
    # will just pass bytestrings around without caring.  We use
    # utf-8 as a reasonable default for common setups.
    encoding = 'utf-8'
    eol = '\n'
    session_factory = None

    logger = logging.getLogger(__name__)

    def __init__(self, channel):
        self.channel = channel
        self.channel.proto = self
        self.reset()

    def reset(self):
        self._state = 'init'
        self.env = {}
        self.argv = []
        self._resp_code = None
        self._resp_message = ''
        self._commands = collections.deque()
        self._session = None

    def connection_made(self, transport):
        self.transport = transport

    def connection_lost(self, exc):
        pass

    def write(self, data):
        self.transport.write(data)

    def bind_session(self):
        """
        """
        if self.session_factory is not None:
            self._session = self.session_factory()
            self._session.proto = self
        return self._session

    def unbind_session(self):
        """
        """
        if self._session is not None:
            self._session.session_finished()
            self._session = None

    def _split_key_value(self, line):
        key, sep, value = line.rstrip().partition(':')
        if not sep:
            raise ValueError("Expected a key/value pair, got %r"
                             % (line,))
        return key, value.lstrip()

    def _escape_arg(self, arg):
        if '\0' in arg or '\n' in arg:
            raise ValueError("Forbidden characters in AGI argument: %r"
                             % (arg,))
        escaped = re.sub(r'([\\"])', r'\\\1', arg)
        if not arg or escaped != arg or ' ' in arg or '\t' in arg:
            return '"%s"' % escaped
        else:
            return escaped

    def line_received(self, line):
        """
        Processing an incoming *line* of AGI data.
        """
        if not isinstance(line, str):
            # Python 3 only
            line = line.decode(self.encoding)
        self.logger.debug("Incoming line: %r", line)
        if self._state == 'init':
            # In AGI variables header: parse a "key: value" pair
            line = line.rstrip('\r\n')
            if not line:
                # Empty line => AGI variables header finished.
                self._state = 'idle'
                self.logger.info("Got %d AGI variables, now "
                                 "waiting for commands", len(self.env))
                if self._session is not None:
                    self._session.session_established()
                return
            k, v = self._split_key_value(line)
            if not k.startswith('agi_'):
                raise ValueError("Invalid AGI variable %r" % k)
            agi_var = k[4:]
            if agi_var.startswith('arg_'):
                # AGI arguments are passed as 'agi_arg_1', etc.
                try:
                    num_arg = int(agi_var[4:], 10)
                except ValueError:
                    pass
                else:
                    if num_arg == len(self.argv) + 1:
                        self.argv.append(v)
                        return
            if agi_var in self.env:
                self.logger.warning("Duplicate value for AGI variable %r"
                                    % (k,))
            self.env[agi_var] = v
            return
        if self._state == 'idle':
            # AGI is essentially a request-response protocol, the server
            # shouldn't send anything without us asking.
            if line.strip():
                self.logger.warning("Unexpected line received while idle: %r", line)
            return
        if self._state == 'awaiting-response':
            if line[3] not in ' -':
                raise ValueError("Invalid response line %r" % line)
            code = int(line[:3])
            tail = line[4:]
            if not 200 <= code < 600:
                raise ValueError("Invalid response code %d" % code)
            if code < 300:
                self._got_successful_response(code, tail.rstrip())
                return
            if code == 520 and 'follows' in tail:
                # 520 responses can include a multiline usage message
                self._resp_code = code
                self._resp_message = tail
                self._state = 'in-response'
                return
            self._got_error_response(code, tail.rstrip())
            return
        if self._state == 'in-response':
            if line.startswith("%d " % self._resp_code):
                self._got_error_response(self._resp_code,
                                         self._resp_message)
                self._resp_code = None
                self._resp_message = None
            elif line.endswith("520 End of proper usage.\n"):
                # Workaround Async AGI bug where there is a missing EOL
                # inside the "Result" header value.
                self._resp_message += line
                self._got_error_response(self._resp_code,
                                         self._resp_message)
                self._resp_code = None
                self._resp_message = None
            else:
                self._resp_message += line
            return
        # Shouldn't come here
        assert 0

    def send_command(self, args):
        """
        Send the command identified by the *args* tuple of strings.
        """
        # XXX Relax this?
        if self._state != 'idle':
            raise RuntimeError("Can only send AGI command when idle")
        assert not self._commands
        line = self._encode_command(args)
        command = self.channel.send_command_line(line)
        self._state = 'awaiting-response'
        return command

    def _encode_command(self, args):
        """
        Return the encoded command line (as bytes) for the given *args*.
        """
        if len(args) == 0:
            raise ValueError("Args sequence cannot be empty")
        args = [self._escape_arg(a) for a in args]
        line = ' '.join(args) + self.eol
        if not isinstance(line, bytes):
            # Python 3
            line = line.encode(self.encoding)
        return line

    def _parse_result(self, line):
        """
        Parse an AGI result line and return a (result code, variables, data)
        tuple.  For example, given "result=1 (foo bar) endpos=123", return
        (1, {'endpos': '123'), 'foo bar').
        """
        result = None
        variables = {}
        data = None
        data_parts = []
        in_data = False
        for part in line.split(' '):
            if in_data:
                if part.endswith(')'):
                    data_parts.append(part[:-1])
                    in_data = False
                else:
                    data_parts.append(part)
            else:
                if part.startswith('('):
                    if part.endswith(')'):
                        data_parts.append(part[1:-1])
                    else:
                        data_parts.append(part[1:])
                        in_data = True
                else:
                    key, sep, value = part.partition('=')
                    if not sep:
                        # XXX what else?
                        continue
                    if key == 'result':
                        result = int(value, 10)
                    else:
                        variables[key] = value
        return result, variables, ' '.join(data_parts) if data_parts else None

    def _push_command(self, command):
        """
        Push a command on the commands queue and adjust the protocol's state.
        """
        self._commands.append(command)
        self._state = 'awaiting-response'

    def _pop_command(self):
        """
        Pop a command from the commands queue and adjust the protocol's state.
        """
        command = self._commands.popleft()
        self._state = 'awaiting-response' if self._commands else 'idle'
        return command

    def _got_successful_response(self, code, body):
        command = self._pop_command()
        # Try to parse the AGI ad-hoc result string
        result, variables, data = self._parse_result(body)
        if result < 0:
            # A negative result signals failure to execute a command.
            # Unfortunately, Asterisk doesn't give any error description.
            command.set_exception(AGICommandFailure(body))
        else:
            command.set_result(Response(result, variables, data))

    def _got_error_response(self, code, message):
        command = self._pop_command()
        exc_class = _agi_errors.get(code, AGIError)
        exc = exc_class(message)
        command.set_exception(exc)




import hashlib
import logging
import os

from obelus.ami.protocol import Handler, ActionError


log = logging.getLogger(__name__)


class OriginateError(ActionError):
    """
    Error triggered by an OriginateResponse of type "Failure".
    """
    # XXX see e.g. http://permalink.gmane.org/gmane.comp.telephony.pbx.asterisk.user/210539
    # for the meaning of the reason.

    def __init__(self, reason):
        self.reason = int(reason, 10)

    def __str__(self):
        return "Originate failed with reason %s" % (self.reason)


class Call(object):
    """
    Base class for call objects tracked by the CallManager.
    It is recommended you subclass this class and override the various
    event handlers.
    """

    _call_id = None
    _action_id = None
    manager = None

    def _bind(self, manager, call_id):
        self.manager = manager
        self._call_id = call_id
        self._state = None
        self._state_desc = None
        self._unique_ids = set()

    def __str__(self):
        try:
            return "Call #%s" % self._call_id
        except AttributeError:
            return super(Call, self).__str__()

    def unique_ids(self):
        """
        Get the alive channels related this call.
        Return a list of unique ids.
        """
        try:
            return sorted(self._unique_ids)
        except AttributeError:
            pass
        raise ValueError("Call not originated")

    # XXX should the state notification callbacks get the logical
    # channel number? (i.e. 1 for the first created channel, 2 for the
    # second...)

    def call_queued(self):
        """
        Called when a call is queued (i.e. when the Originate command
        is accepted).
        """

    def call_failed(self, exc):
        """
        Called when a call fails early.  This either means the Originate
        command was rejected, or a failed OriginateResponse event was
        received just after the command was accepted.
        """

    def call_state_changed(self, state, state_desc):
        """
        Called when a call has its state changed.  *state* is the new
        numeric state, *state_desc* its textual description.
        """

    def dialing_started(self):
        """
        Called when dialing the call has started.
        """

    def dialing_finished(self, status):
        """
        Called when dialing the call has finished with the given *status*.
        """

    def call_ended(self, cause, cause_desc):
        """
        Called when the call ends.  *cause* is the numeric cause
        sent by Asterisk, *cause_desc* its textual description.
        """


class CallManager(object):
    """
    A CallManager helps you originate calls and track the status of those
    calls using an AMI instance.
    """
    # Implementation strategy: when originating a call, we add a custom
    # variable with a unique name and a unique per-call value.  We then
    # listen to SetVar events, which allows us to associate a given
    # call with the first channel allocated for it.
    # Another strategy would be to hijack the CallerID field for our
    # unique id, but this would prevent the user from passing important
    # information there.

    def __init__(self, ami):
        self.ami = ami
        self._tracking_variable = (
            'X_' + hashlib.sha1(os.urandom(32)).hexdigest().upper()[:12])
        self._call_id = 1
        # action id => call (queued but untracked calls)
        self._actions = {}
        # call id => call (all queued calls)
        self._calls = {}
        # channel unique id => call
        self._unique_ids = {}
        self.setup_event_handlers()

    def _new_call_id(self):
        i = self._call_id
        self._call_id = i + 1
        return str(i)

    def queued_calls(self):
        """
        Return a set of all queued calls.
        """
        return set(self._calls.values())

    def tracked_calls(self):
        """
        Return a set of currently tracked calls (i.e. queued *and*
        successfully originated).  This is a subset of queued_calls().
        """
        return set(self._calls.values()) - set(self._actions.values())

    def setup_event_handlers(self):
        """
        Setup the AMI event handlers required for call tracking.
        This is implicitly called on __init__().
        """
        self.ami.register_event_handler('VarSet', self.on_var_set)
        self.ami.register_event_handler('LocalBridge', self.on_local_bridge)
        self.ami.register_event_handler('Dial', self.on_dial)
        self.ami.register_event_handler('Newstate', self.on_new_state)
        self.ami.register_event_handler('Hangup', self.on_hangup)
        # Yes, there's an event called "OriginateResponse"
        self.ami.register_event_handler('OriginateResponse', self.on_originate_response)

    def setup_filters(self):
        """
        Setup server-side AMI event filters tailored for this CallManager,
        in order to limit resource consumption in the AMI protocol.

        Calling this method is not required for proper functioning, but
        recommended if your Asterisk receives a lot of traffic and generates
        a lot of AMI events by default.

        This method only relies on whitelisting, you can therefore
        setup more server-side filters if you are interested in other
        events.
        """
        # We are interesting in "call" events, as well as in those
        # events that mention our tracking variable (normally, it's
        # only one SetVar event per successfully originated call).
        # This spares us the bursts of NewExten, VarSet and AGIExec events
        # that can occur on non-trivial Asterisk setups.
        filters = ['Privilege: call,all',
                   'Variable: ' + self._tracking_variable]
        handlers = [self.ami.send_action('Filter',
                                         {'Operation': 'Add',
                                          'Filter': filter})
                    for filter in filters]
        return Handler.aggregate(handlers)

    def originate(self, call, headers, variables=None):
        """
        Originate a *call* with the given *headers* (and, optionally,
        call-specific *variables*).  *call* should be a Call instance.
        """
        if not isinstance(call, Call):
            raise TypeError("expected a Call instance, got %r"
                            % call.__class__)
        if call.manager is not None:
            raise ValueError("cannot reuse Call instance, need a new one")
        call_id = self._new_call_id()
        variables = variables or {}
        variables[self._tracking_variable] = call_id
        a = self.ami.send_action('Originate', headers, variables)
        call._bind(self, call_id)
        def _call_queued(resp):
            action_id = resp.headers['ActionID']
            call._action_id = action_id
            self._actions[action_id] = call
            self._calls[call_id] = call
            call.call_queued()
        def _call_failed(exc):
            call.call_failed(exc)
        a.on_result = _call_queued
        a.on_exception = _call_failed

    def on_originate_response(self, event):
        h = event.headers
        if h['Response'] != 'Failure':
            return
        action_id = h['ActionID']
        call = self._actions.pop(action_id, None)
        if call is not None:
            del self._calls[call._call_id]
            call.call_failed(OriginateError(h['Reason']))

    def on_var_set(self, event):
        h = event.headers
        if h['Variable'] != self._tracking_variable:
            return
        call_id = h['Value']
        unique_id = h['Uniqueid']
        try:
            call = self._calls[call_id]
        except KeyError:
            log.error("Got unknown call_id in SetVar: %s", call_id)
            return
        try:
            del self._actions[call._action_id]
        except KeyError:
            log.error("Got duplicate SetVar for call #%s", call_id)
            return
        log.info("Got UniqueID %r for call #%s (channel %r)",
                 unique_id, call_id, h['Channel'])
        call._unique_ids.add(unique_id)
        self._unique_ids[unique_id] = call

    def on_local_bridge(self, event):
        h = event.headers
        id1 = h['Uniqueid1']
        call = self._unique_ids.get(id1)
        if call is None:
            log.debug("LocalBridge: unknown UniqueID %r, ignoring", id1)
            return
        id2 = h['Uniqueid2']
        other_call = self._unique_ids.get(id2)
        if other_call is not None and other_call is not call:
            log.error("LocalBridge: UniqueID %r already bound to other call #%s",
                      id2, other_call._call_id)
            return
        log.info("LocalBridge: new related UniqueID %r for call #%s",
                 id2, call._call_id)
        call._unique_ids.add(id2)
        self._unique_ids[id2] = call

    def on_hangup(self, event):
        h = event.headers
        unique_id = h['Uniqueid']
        call = self._unique_ids.pop(unique_id, None)
        if call is None:
            log.debug("Hangup: unknown UniqueID %r, ignoring", unique_id)
            return
        call._unique_ids.remove(unique_id)
        if not call._unique_ids:
            del self._calls[call._call_id]
            call.call_ended(int(h['Cause'], 10), h.get('Cause-txt', ''))

    def on_dial(self, event):
        h = event.headers
        unique_id = h['UniqueID']  # casing!
        call = self._unique_ids.get(unique_id)
        if call is None:
            log.debug("Dial: unknown UniqueID %r, ignoring", unique_id)
            return
        sub = h['SubEvent']
        if sub == 'Begin':
            call.dialing_started()
        elif sub == 'End':
            status = h['DialStatus']
            call.dialing_finished(status)

    def on_new_state(self, event):
        h = event.headers
        unique_id = h['Uniqueid']
        call = self._unique_ids.get(unique_id)
        if call is None:
            log.debug("Newstate: unknown UniqueID %r, ignoring", unique_id)
            return
        state = int(h['ChannelState'])
        state_desc = h['ChannelStateDesc']
        if state != call._state:
            call._state = state
            call.call_state_changed(state, state_desc)
        call._state_desc = state_desc



"""
AMI adapter for the Tornado network programming framework.
"""

try:
    import tornado
except ImportError:
    tornado = None

if not tornado:
    raise ImportError("tornado is required for this module to work: "
                      "http://www.tornadoweb.org/")

from ..tornadosupport import TornadoAdapter


if __name__ == "__main__":
    import logging
    import socket

    from tornado.ioloop import IOLoop
    from tornado.iostream import IOStream

    from . import examplecli

    parser = examplecli.create_parser(
        description="Tornado-based AMI client example")

    options, args = examplecli.parse_args(parser)

    log = logging.getLogger(__name__)

    loop = IOLoop.instance()
    proto = examplecli.CLIProtocol(loop, options)
    adapter = TornadoAdapter(proto)

    stream = IOStream(socket.socket(), loop)
    stream.connect((options.host, options.port),
                   lambda: adapter.bind_stream(stream))

    try:
        loop.start()
    except KeyboardInterrupt:
        pass




from collections import MutableMapping


_sentinel = object()


class CaseDict(MutableMapping):
    """
    A case-insensitive dictionary.
    """

    __slots__ = ('_data',)

    def __init__(self, __dict=None, **kwargs):
        # lower-cased => (original casing, value)
        self._data = {}
        if __dict is not None:
            self.update(__dict)
        if kwargs:
            self.update(kwargs)

    # Minimum set of methods required for MutableMapping

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return (v[0] for v in self._data.values())

    def __getitem__(self, key):
        return self._data[key.lower()][1]

    def __setitem__(self, key, value):
        self._data[key.lower()] = (key, value)

    def __delitem__(self, key):
        del self._data[key.lower()]

    # Methods overriden to mitigate the performance overhead.

    def __contains__(self, key):
        return key.lower() in self._data

    def clear(self):
        self._data.clear()

    def get(self, key, default=_sentinel):
        tup = self._data.get(key.lower())
        if tup is not None:
            return tup[1]
        elif default is not _sentinel:
            return default
        else:
            return None

    def pop(self, key, default=_sentinel):
        if default is not _sentinel:
            tup = self._data.pop(key.lower(), default)
        else:
            tup = self._data.pop(key.lower())
        if tup is not default:
            return tup[1]
        else:
            return default

    # Other methods

    def __repr__(self):
        if self._data:
            return '%s(%r)' % (self.__class__.__name__, dict(self))
        else:
            return '%s()' % (self.__class__.__name__)




from functools import partial
import re
import unittest

from mock import Mock, ANY

from obelus.ami.calls import Call, CallManager, OriginateError
from obelus.ami.protocol import (
    BaseAMIProtocol, AMIProtocol, Event, Response, EventList, ActionError)
from obelus.test import main, watch_logging
from obelus.test.test_amiprotocol import ProtocolTestBase


class MockCall(Call):

    def __init__(self):
        self.event_calls = []
        for meth_name in [
            'call_queued',
            'call_failed',
            'call_state_changed',
            'dialing_started',
            'dialing_finished',
            'call_ended']:
            def side_effect(meth_name=meth_name):
                def mocked(*args):
                    self.event_calls.append(meth_name)
                return mocked
            setattr(self, meth_name, Mock(side_effect=side_effect()))


UNIQUE_ID = '1378719573.625'
CHANNEL = 'Local/6004@default-00000118;1'
UNIQUE_ID_2 = '1378719573.626'
CHANNEL_2 = 'Local/6004@default-00000118;2'
UNIQUE_ID_OTHER = '1378719573.631'
UNIQUE_ID_OTHER_2 = '1378719573.632'

LOCAL_BRIDGE = Event('LocalBridge',
                     {'Uniqueid2': UNIQUE_ID_2,
                      'Uniqueid1': UNIQUE_ID,
                      'Channel2': CHANNEL_2,
                      'Channel1': CHANNEL,
                      'Context': 'default'})

LOCAL_BRIDGE_OTHER = Event('LocalBridge',
                     {'Uniqueid2': UNIQUE_ID_OTHER_2,
                      'Uniqueid1': UNIQUE_ID_OTHER,
                      'Channel2': 'Local/6004@default-00000119;2',
                      'Channel1': 'Local/6004@default-00000119;1',
                      'Context': 'default'})

DIAL_START = Event('Dial',
                   {'DestUniqueID': '1378719683.629',
                    'SubEvent': 'Begin',
                    'UniqueID': UNIQUE_ID_2,
                    'Channel': CHANNEL_2,
                    'Dialstring': '0sxqaw'})

DIAL_END = Event('Dial',
                 {'SubEvent': 'End',
                  'UniqueID': UNIQUE_ID_2,
                  'Channel': CHANNEL_2,
                  'DialStatus': 'BUSY'})

DIAL_OTHER = Event('Dial',
                   {'DestUniqueID': '1378719683.629',
                    'SubEvent': 'Begin',
                    'UniqueID': UNIQUE_ID_OTHER,
                    'Channel': CHANNEL_2,
                    'Dialstring': '0sxqaw'})

NEWSTATE_1_OTHER = Event('Newstate',
                         {'ChannelState': '5',
                          'Uniqueid': '1378719683.629',
                          'Channel': 'SIP/0sxqaw-00000041',
                          'ChannelStateDesc': 'Ringing'})

NEWSTATE_1 = Event('Newstate',
                   {'ChannelState': '5',
                    'Uniqueid': UNIQUE_ID,
                    'Channel': CHANNEL,
                    'ChannelStateDesc': 'Ringing'})

NEWSTATE_2 = Event('Newstate',
                   {'ChannelState': '6',
                    'Uniqueid': UNIQUE_ID_2,
                    'Channel': CHANNEL_2,
                    'ChannelStateDesc': 'Up'})

HANGUP_REJECTED_1 = Event('Hangup',
                          {'Cause-txt': 'Call Rejected',
                           'Uniqueid': UNIQUE_ID_2,
                           'Cause': '21',
                           'Channel': CHANNEL_2})

HANGUP_REJECTED_2 = Event('Hangup',
                          {'Cause-txt': 'Call Rejected',
                           'Uniqueid': UNIQUE_ID,
                           'Cause': '21',
                           'Channel': CHANNEL})


class CallManagerTest(ProtocolTestBase, unittest.TestCase):

    protocol_factory = AMIProtocol

    def call(self):
        return MockCall()

    def call_manager(self):
        self.ami = self.ready_proto()
        cm = CallManager(self.ami)
        cm._tracking_variable = 'X_TRACK'
        return cm

    def queued_call(self):
        cm = self.call_manager()
        call = self.call()
        cm.ami._action_id = 1
        cm.ami.write = Mock()
        cm.originate(call, {"Foo": "Bar"})
        resp = Response('success', {'ActionID': '1'}, [])
        cm.ami.response_received(resp)
        return cm, call

    def tracked_call(self):
        cm, call = self.queued_call()
        event = Event('VarSet',
                      {'Variable': 'X_TRACK',
                       'Value': '1',
                       'Channel': CHANNEL,
                       'Uniqueid': UNIQUE_ID})
        cm.ami.event_received(event)
        return cm, call

    def test_init(self):
        ami = self.protocol_factory()
        cm = CallManager(ami)
        self.assertEqual(cm.queued_calls(), set())
        self.assertEqual(cm.tracked_calls(), set())
        self.assertIs(cm.ami, ami)

    def test_tracking_variable(self):
        cms = [CallManager(self.protocol_factory()) for i in range(5)]
        tracking_variables = {cm._tracking_variable for cm in cms}
        self.assertEqual(len(tracking_variables), len(cms), tracking_variables)
        one_var = tracking_variables.pop()
        self.assertTrue(re.match("^[A-Z][_A-Z0-9]*$", one_var), one_var)

    def test_new_call_id(self):
        cm = self.call_manager()
        self.assertEqual(cm._new_call_id(), "1")
        self.assertEqual(cm._new_call_id(), "2")
        self.assertEqual(cm._new_call_id(), "3")

    def test_setup_filters(self):
        cm = self.call_manager()
        sa = cm.ami.send_action = Mock()
        cm.setup_filters()
        self.assertEqual(sa.call_count, 2)
        for (c, filter_desc) in zip(sa.call_args_list,
                                    ["Privilege: call,all",
                                     "Variable: X_TRACK"]):
            (name, headers), _ = c
            self.assertEqual(name, 'Filter')
            self.assertEqual(headers, {'Operation': 'Add',
                                       'Filter': filter_desc})

    def test_originate_bad_call_type(self):
        cm = self.call_manager()
        with self.assertRaises(TypeError):
            cm.originate(object(), {})
        with self.assertRaises(TypeError):
            cm.originate("foo", {})
        self.assertEqual(cm.queued_calls(), set())
        self.assertEqual(cm.tracked_calls(), set())

    def test_originate(self):
        cm = self.call_manager()
        call = self.call()
        sa = cm.ami.send_action = Mock()
        cm.originate(call, {"Foo": "Bar"})
        sa.assert_called_once_with(
            "Originate", {"Foo": "Bar"}, {"X_TRACK": "1"})
        self.assertIs(call.manager, cm)
        # The call is still not queued: we wait for the AMI's response
        self.assertEqual(cm.queued_calls(), set())
        self.assertEqual(cm.tracked_calls(), set())
        self.assertEqual(call.event_calls, [])

    def test_originate_call_reuse(self):
        cm = self.call_manager()
        call = self.call()
        sa = cm.ami.send_action = Mock()
        cm.originate(call, {"Foo": "Bar"})
        with self.assertRaises(ValueError):
            cm.originate(call, {"Foo": "Quux"})
        self.assertEqual(sa.call_count, 1)

    def test_sync_originate_failure(self):
        # Synchronous failure: the originate action's response is a failure
        cm = self.call_manager()
        call = self.call()
        cm.ami._action_id = 1
        cm.ami.write = Mock()
        cm.originate(call, {"Foo": "Bar"})
        resp = Response('error', {'Message': 'Extension does not exist.',
                                  'ActionID': '1'}, [])
        cm.ami.response_received(resp)
        self.assertEqual(call.event_calls, ['call_failed'])
        self.assert_called_once_with_exc(call.call_failed, ActionError)

    def test_call_originated_unique_ids(self):
        cm = self.call_manager()
        call = self.call()
        cm.ami._action_id = 1
        cm.ami.write = Mock()
        self.assertRaises(ValueError, call.unique_ids)
        cm.originate(call, {"Foo": "Bar"})
        self.assertEqual(call.unique_ids(), [])

    def test_call_queued_unique_ids(self):
        cm, call = self.queued_call()
        self.assertEqual(call.unique_ids(), [])

    def test_call_tracked_unique_ids(self):
        cm, call = self.tracked_call()
        self.assertEqual(call.unique_ids(), [UNIQUE_ID])

    def test_local_bridges(self):
        cm, call = self.tracked_call()
        self.assertEqual(call.unique_ids(), [UNIQUE_ID])
        cm.ami.event_received(LOCAL_BRIDGE_OTHER)
        self.assertEqual(call.unique_ids(), [UNIQUE_ID])
        cm.ami.event_received(LOCAL_BRIDGE)
        self.assertEqual(call.unique_ids(), [UNIQUE_ID, UNIQUE_ID_2])

    def test_call_queued(self):
        cm, call = self.queued_call()
        self.assertEqual(call.event_calls, ['call_queued'])
        call.call_queued.assert_called_once_with()
        self.assertEqual(cm.queued_calls(), {call})
        self.assertEqual(cm.tracked_calls(), set())

    def test_async_originate_failure(self):
        # Early asynchronous failure: the originate action's response is
        # a success, but a failed OriginateResponse event comes just after
        cm, call = self.queued_call()
        event = Event('OriginateResponse',
                      {'Uniqueid': '<null>',
                       'Reason': '0',
                       'ActionID': '1',
                       'Response': 'Failure'})
        cm.ami.event_received(event)
        self.assertEqual(call.event_calls, ['call_queued', 'call_failed'])
        exc = self.assert_called_once_with_exc(call.call_failed, OriginateError)
        self.assertEqual(exc.reason, 0)

    def test_channel_matching(self):
        cm, call = self.queued_call()
        # This one is ignored
        event = Event('Newchannel',
                      {'Uniqueid': UNIQUE_ID,
                       'ChannelState': '0',
                       'Channel': CHANNEL,
                       'ChannelStateDesc': 'Down'})
        cm.ami.event_received(event)
        self.assertEqual(cm.tracked_calls(), set())
        # Not the right variable name => ignored
        event = Event('VarSet',
                      {'Variable': 'SOMETHING',
                       'Value': '1',
                       'Channel': CHANNEL,
                       'Uniqueid': UNIQUE_ID})
        cm.ami.event_received(event)
        self.assertEqual(cm.tracked_calls(), set())
        # Not the right variable value => ignored too
        event = Event('VarSet',
                      {'Variable': 'X_TRACK',
                       'Value': '42',
                       'Channel': CHANNEL,
                       'Uniqueid': UNIQUE_ID})
        with watch_logging('obelus.ami.calls', level='WARN') as w:
            cm.ami.event_received(event)
        self.assertEqual(cm.tracked_calls(), set())
        self.assertEqual(len(w.output), 1)
        # This one sets up the matching
        event = Event('VarSet',
                      {'Variable': 'X_TRACK',
                       'Value': '1',
                       'Channel': CHANNEL,
                       'Uniqueid': UNIQUE_ID})
        cm.ami.event_received(event)
        self.assertEqual(cm.queued_calls(), {call})
        self.assertEqual(cm.tracked_calls(), {call})

    def test_early_hangup(self):
        # Test getting a Hangup without an earlier LocalBridge
        cm, call = self.tracked_call()
        # Ignored (unknown channel even though it's probably related)
        event = Event('Hangup',
                      {'Cause-txt': 'Unknown',
                       'Uniqueid': UNIQUE_ID_2,
                       'Cause': '0',
                       'Channel': CHANNEL_2})
        cm.ami.event_received(event)
        self.assertEqual(call.event_calls, ['call_queued'])
        # This one triggers the call's end
        event = Event('Hangup',
                      {'Cause-txt': 'Unknown',
                       'Uniqueid': UNIQUE_ID,
                       'Cause': '0',
                       'Channel': CHANNEL})
        cm.ami.event_received(event)
        self.assertEqual(call.event_calls, ['call_queued', 'call_ended'])
        call.call_ended.assert_called_once_with(0, 'Unknown')

    def test_dialing_started(self):
        cm, call = self.tracked_call()
        # The LocalBridge events helps track the second channel, on which
        # the dialing will happen.
        cm.ami.event_received(LOCAL_BRIDGE)
        self.assertEqual(call.event_calls, ['call_queued'])
        cm.ami.event_received(DIAL_START)
        self.assertEqual(call.event_calls, ['call_queued', 'dialing_started'])
        call.dialing_started.assert_called_once_with()

    def test_dialing_finished(self):
        cm, call = self.tracked_call()
        cm.ami.event_received(LOCAL_BRIDGE)
        cm.ami.event_received(DIAL_START)
        cm.ami.event_received(DIAL_END)
        self.assertEqual(call.event_calls, ['call_queued', 'dialing_started',
                                            'dialing_finished'])
        call.dialing_finished.assert_called_once_with('BUSY')

    def test_dial_other(self):
        cm, call = self.tracked_call()
        cm.ami.event_received(LOCAL_BRIDGE)
        cm.ami.event_received(DIAL_OTHER)
        self.assertEqual(call.event_calls, ['call_queued'])

    def test_state_events(self):
        cm, call = self.tracked_call()
        cm.ami.event_received(LOCAL_BRIDGE)
        cm.ami.event_received(DIAL_START)
        cm.ami.event_received(NEWSTATE_1_OTHER)
        self.assertEqual(call.event_calls, ['call_queued', 'dialing_started'])
        cm.ami.event_received(NEWSTATE_1)
        self.assertEqual(call.event_calls,
                         ['call_queued', 'dialing_started', 'call_state_changed'])
        call.call_state_changed.assert_called_once_with(5, 'Ringing')
        # call_state_changed not called again if the state doesn't change
        cm.ami.event_received(NEWSTATE_1)
        call.call_state_changed.assert_called_once_with(5, 'Ringing')
        cm.ami.event_received(NEWSTATE_2)
        self.assertEqual(call.event_calls,
                         ['call_queued', 'dialing_started',
                          'call_state_changed', 'call_state_changed'])
        call.call_state_changed.assert_called_with(6, 'Up')

    def test_call_ended(self):
        cm, call = self.tracked_call()
        cm.ami.event_received(LOCAL_BRIDGE)
        cm.ami.event_received(HANGUP_REJECTED_1)
        self.assertEqual(call.event_calls, ['call_queued'])
        cm.ami.event_received(HANGUP_REJECTED_2)
        self.assertEqual(call.event_calls, ['call_queued', 'call_ended'])
        call.call_ended.assert_called_once_with(21, 'Call Rejected')

    def test_call_ended_2(self):
        # Same as test_call_ended(), but Hangups received in reverse order.
        cm, call = self.tracked_call()
        cm.ami.event_received(LOCAL_BRIDGE)
        cm.ami.event_received(HANGUP_REJECTED_2)
        self.assertEqual(call.event_calls, ['call_queued'])
        cm.ami.event_received(HANGUP_REJECTED_1)
        self.assertEqual(call.event_calls, ['call_queued', 'call_ended'])
        call.call_ended.assert_called_once_with(21, 'Call Rejected')


if __name__ == "__main__":
    main()




import unittest

from obelus.common import LineReceiver
from obelus.test import main


class MockLineReceiver(LineReceiver):
    def __init__(self):
        super(MockLineReceiver, self).__init__()
        self.lines = []

    def line_received(self, line):
        self.lines.append(line)


class LineReceiverTest(unittest.TestCase):
    DATA = b"one line\nanother line\r\nyet another\rtail\n"
    EXPECTED = [b"one line\n", b"another line\r\n", b"yet another\r", b"tail\n"]
    # This can happen if the '\r\n' is split between two data_received()
    # calls.
    EXPECTED_ALT = [b"one line\n", b"another line\r", b"yet another\r", b"tail\n"]

    def test_receive_large(self):
        lr = MockLineReceiver()
        lr.data_received(self.DATA)
        self.assertEqual(lr.lines, self.EXPECTED)

    def test_receive_bytewise(self):
        lr = MockLineReceiver()
        for i in range(len(self.DATA)):
            lr.data_received(self.DATA[i:i + 1])
        self.assertIn(lr.lines, (self.EXPECTED, self.EXPECTED_ALT))

    def test_receive_chunkwise(self):
        for chunk_size in range(2, 20):
            chunks = [self.DATA[i:i + chunk_size]
                      for i in range(0, len(self.DATA), chunk_size)]
            lr = MockLineReceiver()
            for chunk in chunks:
                lr.data_received(chunk)
            self.assertIn(lr.lines, (self.EXPECTED, self.EXPECTED_ALT))

    def test_receive_cr_ending_lines(self):
        # '\r'-ending lines do not wait for the next byte to
        # come before they are notified.
        lr = MockLineReceiver()
        lr.data_received(b'foo\r')
        self.assertEqual(lr.lines, [b'foo\r'])
        lr.lines = []
        # '\n' gets eaten because of previous '\r'
        lr.data_received(b'\n')
        self.assertEqual(lr.lines, [])
        lr.data_received(b'\n\r')
        self.assertEqual(lr.lines, [b'\n', b'\r'])
        lr.lines = []
        # First '\n' gets eaten because of previous '\r'
        lr.data_received(b'\n\na\r\nb\r')
        self.assertEqual(lr.lines, [b'\n', b'a\r\n', b'b\r'])
        lr.lines = []
        lr.data_received(b'\rc\n')
        self.assertEqual(lr.lines, [b'\r', b'c\n'])


if __name__ == "__main__":
    main()



import argparse
import logging
import pprint
import sys

from .protocol import AGIProtocol
from .session import AGISession

log = logging.getLogger(__name__)


class CLIOptions(object):
    pass


class CLISession(AGISession):

    def session_established(self):
        log.info("AGI session established")
        log.info("AGI variables: %s", pprint.pformat(self.proto.env))
        self.run_coroutine(self.example_commands())

    def session_finished(self):
        self.logger.info("AGI session finished")

    def example_commands(self):
        try:
            resp = yield self.proto.send_command(["channel", "status"])
            log.info("Channel status: %d", resp.result)
            resp = yield self.proto.send_command(["hangup"])
            log.info("Hangup result: %d", resp.result)
        finally:
            self.proto.transport.close()


class CLIProtocol(AGIProtocol):

    session_factory = CLISession

    def connection_made(self, transport=None):
        log.info("Connection made")
        super(CLIProtocol, self).connection_made(transport)

    def connection_lost(self, exc):
        log.info("Connection lost")
        super(CLIProtocol, self).connection_lost(exc)


def create_parser(description):
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-q', '--quiet', action='store_true',
                        help='hide information messages')
    parser.add_argument('--debug', action='store_true',
                        help='show debug messages')
    return parser


def parse_args(parser):
    args = parser.parse_args()
    options = CLIOptions()

    if args.quiet:
        level = logging.WARNING
    elif args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(stream=sys.stderr, level=level)

    return options, args



# -*- coding: utf-8 -*-

import contextlib
import textwrap
import unittest

from mock import Mock, ANY

from obelus.agi.asyncagi import AsyncAGIExecutor
from obelus.agi.protocol import (
    AGIProtocol, Response,
    AGICommandFailure, AGIUnknownCommand, AGIForbiddenCommand, AGISyntaxError)
from obelus.agi.session import AGISession
from obelus.ami.protocol import AMIProtocol, ActionError
from obelus.common import Handler
from obelus.test import main, watch_logging


def literal_ami(text):
    text = textwrap.dedent(text.rstrip() + '\n\n')
    if not isinstance(text, bytes):
        text = text.encode('utf-8')
    return text

AMI_GREETING_LINE = b"Asterisk Call Manager/1.4\r\n"

ASYNC_AGI_START = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: Start
    Channel: Local/678@default-00000012;2
    Env: agi_request%3A%20async%0Aagi_channel%3A%20Local%2F678%40default-00000012%3B2%0Aagi_language%3A%20en%0Aagi_type%3A%20Local%0Aagi_uniqueid%3A%201377871527.46%0Aagi_version%3A%2011.5.0%2Bpf.xivo.13.16~20130722.141054.2668289%0Aagi_callerid%3A%20unknown%0Aagi_calleridname%3A%20unknown%0Aagi_callingpres%3A%2067%0Aagi_callingani2%3A%200%0Aagi_callington%3A%200%0Aagi_callingtns%3A%200%0Aagi_dnid%3A%20unknown%0Aagi_rdnis%3A%20unknown%0Aagi_context%3A%20default%0Aagi_extension%3A%20678%0Aagi_priority%3A%201%0Aagi_enhanced%3A%200.0%0Aagi_accountcode%3A%20%0Aagi_threadid%3A%20-1257751696%0A%0A
    """)

# Non-ASCII 'Channel', Non-ASCII 'Env'
ASYNC_AGI_START_NON_ASCII = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: Start
    Channel: Local/678é@default-00000008;2
    Env: agi_request%3A%20async%0Aagi_channel%3A%20Local%2F678%C3%A9%40default-00000008%3B2%0Aagi_language%3A%20en%0Aagi_type%3A%20Local%0Aagi_uniqueid%3A%201378115103.17%0Aagi_version%3A%2011.5.0%2Bpf.xivo.13.16~20130722.141054.2668289%0Aagi_callerid%3A%20unknown%0Aagi_calleridname%3A%20unknown%0Aagi_callingpres%3A%2067%0Aagi_callingani2%3A%200%0Aagi_callington%3A%200%0Aagi_callingtns%3A%200%0Aagi_dnid%3A%20unknown%0Aagi_rdnis%3A%20unknown%0Aagi_context%3A%20default%0Aagi_extension%3A%20678%C3%A9%0Aagi_priority%3A%201%0Aagi_enhanced%3A%200.0%0Aagi_accountcode%3A%20%0Aagi_threadid%3A%20-1258722448%0A%0A
    """)

# 'Env' lacks the trailing double-EOL
ASYNC_AGI_START_INCOMPLETE = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: Start
    Channel: Local/678@default-00000012;2
    Env: agi_request%3A%20async%0A
    """)

ASYNC_AGI_END = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: End
    Channel: Local/678@default-00000012;2
    """)

OTHER_ASYNC_AGI_END = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: End
    Channel: Local/some-unknown-channel
    """)

AGI_ACTION_SUCCESS = literal_ami("""\
    Response: Success
    ActionID: 1
    Message: Added AGI command to queue
    """)

AGI_ACTION_ERROR = literal_ami("""\
    Response: Error
    ActionID: 1
    Message: Channel does not exist.
    """)

# '200 result=0 (foobar) endpos=1234'
ASYNC_AGI_EXEC_1 = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: Exec
    Channel: Local/678@default-00000012;2
    CommandID: SOME-COMMAND-ID
    Result: 200%20result%3D0%20%28foobar%29%20endpos%3D1234%0A
    """)

ASYNC_AGI_RESP_1 = Response(result=0, variables={'endpos': '1234'},
                            data='foobar')

# '510 Invalid or unknown command'
ASYNC_AGI_EXEC_2 = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: Exec
    Channel: Local/678@default-00000012;2
    CommandID: SOME-COMMAND-ID
    Result: 510%20Invalid%20or%20unknown%20command%0A
    """)

# '520-Invalid command syntax' + continuation
# NOTE: the "Result" value is malformed, there is a missing EOL
# between the usage string and the "520 End of proper usage" line.
ASYNC_AGI_EXEC_3 = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: Exec
    Channel: Local/678@default-00000012;2
    CommandID: SOME-COMMAND-ID
    Result: 520-Invalid%20command%20syntax.%20%20Proper%20usage%20follows%3A%0AReturns%20%3Cliteral%3E0%3C%2Fliteral%3E%20if%20%3Creplaceable%3Evariablename%3C%2Freplaceable%3E%20is%20not%20set.%20Returns%20%3Cliteral%3E1%3C%2Fliteral%3E%20if%20%3Creplaceable%3Evariablename%3C%2Freplaceable%3E%20is%20set%20and%20returns%20the%20variable%20in%20parentheses.%0AExample%20return%20code%3A%20200%20result%3D1%20(testvariable)520%20End%20of%20proper%20usage.%0A
    """)

ASYNC_AGI_EXEC_OTHER_COMMAND = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: Exec
    Channel: Local/678@default-00000012;2
    CommandID: SOME-OTHER-COMMAND-ID
    Result: 200%20result%3D0%20%28foobar%29%20endpos%3D1234%0A
    """)

ASYNC_AGI_EXEC_OTHER_CHANNEL = literal_ami("""\
    Event: AsyncAGI
    Privilege: agi,all
    SubEvent: Exec
    Channel: Local/some-other-channel
    CommandID: SOME-COMMAND-ID
    Result: 200%20result%3D0%20%28foobar%29%20endpos%3D1234%0A
    """)


class TestHelpers(object):

    def setUp(self):
        def session_factory():
            # Stick the latest created session on this test case instance.
            self.session = Mock(spec_set=AGISession)
            return self.session

        class MyAGIProtocol(AGIProtocol):
            pass
        MyAGIProtocol.session_factory = staticmethod(session_factory)
        self.agi_protocol_factory = MyAGIProtocol
        self.ami = AMIProtocol()
        self.executor = self.make_executor()
        self.session = None

    def make_executor(self):
        return AsyncAGIExecutor(self.agi_protocol_factory)

    def bound_executor(self):
        e = self.executor
        self.ami.line_received(AMI_GREETING_LINE)
        e.bind(self.ami)
        return e

    def feed_ami(self, message):
        lines = message.splitlines(True)
        for line in lines:
            self.ami.line_received(line)

    def assert_one_proto(self, expected_channel_id=None):
        e = self.executor
        self.assertEqual(1, len(e._channels))
        (channel_id, channel), = e._channels.items()
        proto = channel.proto
        if expected_channel_id is not None:
            self.assertEqual(channel_id, expected_channel_id)
            self.assertEqual(proto.env['channel'], expected_channel_id)
        self.assertEqual(proto._state, 'idle')
        # A session was bound to the proto
        session = self.session
        self.assertIsInstance(session, AGISession)
        self.assertEqual(session.proto, proto)
        session.session_established.assert_called_once_with()
        return proto

    def assert_called_once_with_exc(self, callback, exc_class):
        callback.assert_called_once_with(ANY)
        (exc,), _ = callback.call_args
        return exc


class AsyncAGITest(TestHelpers, unittest.TestCase):
    """
    General tests for the Async AGI protocol.
    """

    def test_init(self):
        e = self.executor
        self.assertIs(False, e.is_bound())
        self.assertEqual(0, len(e._channels))
        self.assertIs(self.session, None)

    def test_new_command_id(self):
        def assert_no_dups(ids):
            self.assertEqual(sorted(ids), sorted(set(ids)))
        e = self.executor
        f = self.make_executor()
        e_ids = [e._new_command_id() for i in range(10)]
        f_ids = [f._new_command_id() for i in range(10)]
        assert_no_dups(e_ids)
        assert_no_dups(f_ids)
        assert_no_dups(e_ids + f_ids)

    def test_bind(self):
        e = self.executor
        e.bind(self.ami)
        self.assertIs(True, e.is_bound())
        self.assertRaises(ValueError, e.bind, self.ami)
        self.assertIs(True, e.is_bound())
        self.assertEqual(0, len(e._channels))
        self.assertIs(self.session, None)

    def test_unbind(self):
        e = self.executor
        self.assertRaises(ValueError, e.unbind)
        self.assertIs(False, e.is_bound())
        e.bind(self.ami)
        self.assertIs(True, e.is_bound())
        e.unbind()
        self.assertIs(False, e.is_bound())
        self.assertRaises(ValueError, e.unbind)

    def test_async_agi_start(self):
        e = self.bound_executor()
        self.feed_ami(ASYNC_AGI_START)
        proto = self.assert_one_proto("Local/678@default-00000012;2")
        self.assertEqual(proto.argv, [])
        self.assertEqual(proto.env['threadid'], "-1257751696")
        self.assertEqual(self.session.session_finished.call_count, 0)

    def test_async_agi_start_non_ascii(self):
        e = self.bound_executor()
        self.feed_ami(ASYNC_AGI_START_NON_ASCII)
        proto = self.assert_one_proto("Local/678é@default-00000008;2")
        self.assertEqual(proto.argv, [])
        self.assertEqual(proto.env['threadid'], "-1258722448")
        self.assertEqual(self.session.session_finished.call_count, 0)

    def test_async_agi_start_incomplete(self):
        e = self.bound_executor()
        with watch_logging('obelus.agi', level='WARN') as w:
            self.feed_ami(ASYNC_AGI_START_INCOMPLETE)
        self.assertEqual(0, len(e._channels))
        # XXX should it be None?
        self.assertEqual(self.session.session_established.call_count, 0)
        self.assertEqual(self.session.session_finished.call_count, 0)

    def test_async_agi_end(self):
        e = self.bound_executor()
        self.feed_ami(ASYNC_AGI_START)
        self.assertEqual(1, len(e._channels))
        session = self.session
        self.feed_ami(ASYNC_AGI_END)
        self.assertEqual(0, len(e._channels))
        session.session_established.assert_called_once_with()
        session.session_finished.assert_called_once_with()

    def test_async_agi_end_unknown_channel(self):
        e = self.bound_executor()
        self.feed_ami(ASYNC_AGI_START)
        self.assert_one_proto("Local/678@default-00000012;2")
        with watch_logging('obelus.agi', level='WARN') as w:
            self.feed_ami(OTHER_ASYNC_AGI_END)
        self.assert_one_proto("Local/678@default-00000012;2")
        self.assertEqual(len(w.output), 1)
        self.assertIn("Local/some-unknown-channel", w.output[0])
        self.assertEqual(self.session.session_finished.call_count, 0)


class CommandSendingTest(TestHelpers, unittest.TestCase):
    """
    Tests for sending and matching commands over Async AGI.
    """

    def setUp(self):
        TestHelpers.setUp(self)
        e = self.bound_executor()
        self.feed_ami(ASYNC_AGI_START)
        self.proto = self.assert_one_proto()
        self.ami.write = Mock()
        e._new_command_id = Mock(return_value='SOME-COMMAND-ID')

    def test_send_command(self):
        p = self.proto
        e = self.executor
        ami = self.ami
        h = p.send_command(("set", "variable", "foo", "bar quux"))
        self.assertIsInstance(h, Handler)
        ami.write.assert_called_once_with(ANY)
        (data,), _ = ami.write.call_args
        lines = data.splitlines()
        self.assertEqual(len(lines), 6, lines)
        self.assertIn(b"Channel: Local/678@default-00000012;2", lines)
        self.assertIn(b"Action: AGI", lines)
        self.assertIn(b"ActionID: 1", lines)
        self.assertIn(b'Command: set variable foo "bar quux"', lines)
        self.assertIn(b"CommandID: SOME-COMMAND-ID", lines)
        self.assertEqual(lines[-1], b"")

    def queued_command(self, args):
        p = self.proto
        h = p.send_command(["noop"])
        h.on_result = Mock()
        h.on_exception = Mock()
        # A command is only queued when the "AGI" action gets a successful
        # response.
        self.assertEqual(list(p.channel._commands), [])
        self.feed_ami(AGI_ACTION_SUCCESS)
        self.assertEqual(list(p.channel._commands), ['SOME-COMMAND-ID'])
        return p, h

    def test_command_queued(self):
        p, h = self.queued_command(["noop"])
        self.assertEqual(h.on_result.call_count, 0)
        self.assertEqual(h.on_exception.call_count, 0)

    def test_command_queued_successful(self):
        p, h = self.queued_command(["noop"])
        self.feed_ami(ASYNC_AGI_EXEC_1)
        h.on_result.assert_called_once_with(ASYNC_AGI_RESP_1)
        self.assertEqual(h.on_exception.call_count, 0)
        self.assertEqual(p._state, 'idle')
        self.assertEqual(list(p.channel._commands), [])

    def test_command_queued_failed_1(self):
        p, h = self.queued_command(["noop"])
        self.feed_ami(ASYNC_AGI_EXEC_2)
        self.assertEqual(h.on_result.call_count, 0)
        exc = self.assert_called_once_with_exc(h.on_exception, AGIUnknownCommand)
        self.assertEqual(str(exc), "Invalid or unknown command")
        self.assertEqual(p._state, 'idle')
        self.assertEqual(list(p.channel._commands), [])

    def test_command_queued_failed_2(self):
        p, h = self.queued_command(["noop"])
        self.feed_ami(ASYNC_AGI_EXEC_3)
        self.assertEqual(h.on_result.call_count, 0)
        exc = self.assert_called_once_with_exc(h.on_exception, AGISyntaxError)
        self.assertTrue(str(exc).startswith("Invalid command syntax"),
                        str(exc))
        self.assertEqual(p._state, 'idle')
        self.assertEqual(list(p.channel._commands), [])

    def test_command_queuing_failure(self):
        p = self.proto
        h = p.send_command(["noop"])
        h.on_result = Mock()
        h.on_exception = Mock()
        self.feed_ami(AGI_ACTION_ERROR)
        self.assertEqual(list(p.channel._commands), [])
        self.assertEqual(h.on_result.call_count, 0)
        exc = self.assert_called_once_with_exc(h.on_exception, ActionError)
        self.assertEqual(str(exc), "Channel does not exist.")

    def test_other_command_ignored(self):
        p, h = self.queued_command(["noop"])
        with watch_logging('obelus.agi', level='WARN'):
            self.feed_ami(ASYNC_AGI_EXEC_OTHER_COMMAND)
        self.assertEqual(h.on_result.call_count, 0)
        self.assertEqual(h.on_exception.call_count, 0)
        self.assertEqual(list(p.channel._commands), ['SOME-COMMAND-ID'])

    def test_other_channel_ignored(self):
        p, h = self.queued_command(["noop"])
        with watch_logging('obelus.agi', level='WARN'):
            self.feed_ami(ASYNC_AGI_EXEC_OTHER_CHANNEL)
        self.assertEqual(h.on_result.call_count, 0)
        self.assertEqual(h.on_exception.call_count, 0)
        self.assertEqual(list(p.channel._commands), ['SOME-COMMAND-ID'])





import hashlib
import logging
import os
try:
    # Python 3
    from urllib.parse import unquote_to_bytes
except ImportError:
    # Python 2
    from urllib import unquote as unquote_to_bytes

from ..common import Handler
from .protocol import AGIChannel


class _AsyncAGIChannel(AGIChannel):
    """
    """

    def __init__(self, executor):
        self.executor = executor
        # Command ID => Handler
        self._commands = {}

    def send_command_line(self, line):
        return self.executor._send_command_line(self.proto, line)


class AsyncAGIExecutor(object):
    """
    AsyncAGIExecutor dispatches between a single AMI protocol instance
    and an arbitrary number of AGI channels and protocols.

    *protocol_factory* should be a callable returning a AGIProtocol
    instance.  It will be called each time a new Async AGI channel is
    started by Asterisk through the AMI.
    """

    logger = logging.getLogger(__name__)
    ami = None

    def __init__(self, protocol_factory):
        self.protocol_factory = protocol_factory
        # Channel ID => _AsyncAGIChannel
        self._channels = {}
        # Compute a reasonably random stem for command ids
        self._command_id_stem = hashlib.sha1(os.urandom(32)).hexdigest()[:10]
        self._command_id = 1

    def _new_command_id(self):
        command_id = self._command_id
        self._command_id = command_id + 1
        return "%s-%s" % (command_id, self._command_id_stem)

    def _check_bound(self):
        if self.ami is None:
            raise ValueError("Operation on non-bound executor")

    def is_bound(self):
        """
        Whether this executor is bound to an AMI protocol instance.
        """
        return self.ami is not None

    def bind(self, ami):
        """
        Bind this executor to an AMI protocol instance.
        NOTE: only one AsyncAGIExecutor can be bound to an AMI protocol
        instance, at any given time.
        """
        if self.ami is not None:
            raise ValueError("Executor already bound")
        ami.register_event_handler('AsyncAGI', self._asyncagi_event_received)
        self.ami = ami

    def unbind(self):
        """
        Unbind this executor from the AMI protocol instance it is bound to.
        """
        self._check_bound()
        ami = self.ami
        self.ami = None
        ami.unregister_event_handler('AsyncAGI')

    def _decode_agi_data(self, proto, value):
        """
        Decode some AGI protocol data encoded into an AMI header value.
        """
        return unquote_to_bytes(value)

    def _encode_agi_data(self, proto, data):
        """
        Encode some AGI data for sending over the AMI.
        """
        # The AMI protocol accepts str objects as header values, not bytes
        if not isinstance(data, str):
            return data.decode(proto.encoding)
        else:
            return data

    def _send_command_line(self, proto, line):
        self._check_bound()
        channel = proto.channel
        handler = Handler()
        command_id = self._new_command_id()
        headers = {
            'Command': self._encode_agi_data(proto, line.rstrip()),
            'CommandID': command_id,
            'Channel': proto._channel_id,
            }
        # The 'AGI' action has a synchronous response, and then the actual
        # result of the AGI command comes as an AsyncAGI event (subevent
        # 'Exec').
        action_handler = self.ami.send_action('AGI', headers)
        def _on_action_success(resp):
            channel._commands[command_id] = handler
        def _on_action_failure(exc):
            handler.on_exception(exc)
        action_handler.on_result = _on_action_success
        action_handler.on_exception = _on_action_failure
        return handler

    def _asyncagi_event_received(self, event):
        subevent = event.headers['SubEvent']
        if subevent == 'Start':
            self._asyncagi_start(event)
        elif subevent == 'Exec':
            self._asyncagi_exec(event)
        elif subevent == 'End':
            self._asyncagi_end(event)
        else:
            self.logger.warning("Unknown AsyncAGI subevent received: %r")

    def _asyncagi_exec(self, event):
        channel_id = event.headers['Channel']
        command_id = event.headers['CommandID']
        try:
            channel = self._channels[channel_id]
        except KeyError:
            # Could be from a stale session, only make it a warning
            self.logger.warning(
                "Received 'AsyncAGI exec' event for unknown channel %r",
                channel_id)
            return
        try:
            handler = channel._commands.pop(command_id)
        except KeyError:
            self.logger.warning(
                "Received 'AsyncAGI exec' event for unknown command %r "
                "in channel %r", command_id, channel_id)
            return
        # Ensure the AGI protocol is expecting a response for this command.
        proto = channel.proto
        proto._push_command(handler)
        result_block = self._decode_agi_data(proto, event.headers['Result'])
        result_lines = result_block.splitlines(True)
        for line in result_lines:
            proto.line_received(line)
        if proto._state != 'idle':
            self.logger.error(
                "Invalid AGI protocol state after AsyncAGI Exec "
                "(bad 'Env' line?): %r" % (proto._state,))
            return

    def _asyncagi_start(self, event):
        channel_id = event.headers['Channel']
        if channel_id in self._channels:
            self.logger.error(
                "Received new 'AsyncAGI start' event for bound channel %r",
                channel_id)
            return
        channel = _AsyncAGIChannel(self)
        proto = self.protocol_factory(channel)
        proto._channel_id = channel_id
        proto.bind_session()
        # The 'Env' header contains a %-encoded sequence of lines
        # containing the AGI environment, feed it to the protocol.
        env_block = self._decode_agi_data(proto, event.headers['Env'])
        env_lines = env_block.splitlines(True)
        for line in env_lines:
            proto.line_received(line)
        if proto._state != 'idle':
            self.logger.error(
                "Invalid AGI protocol state after AsyncAGI Start "
                "(bad 'Env' line?): %r" % (proto._state,))
            return
        self._channels[channel_id] = channel

    def _asyncagi_end(self, event):
        # TODO error out all the pending commands
        channel_id = event.headers['Channel']
        try:
            channel = self._channels.pop(channel_id)
        except KeyError:
            # Could be from a stale session, only make it a warning
            self.logger.warning(
                "Received 'AsyncAGI end' event for unknown channel %r",
                channel_id)
            return
        channel.proto.unbind_session()
        del channel.proto



# -*- coding: utf-8 -*-

import contextlib
import textwrap
import unittest

from mock import Mock, ANY

from obelus.agi.protocol import (
    AGIProtocol, ProtocolAGIChannel, Response,
    AGICommandFailure, AGIUnknownCommand, AGIForbiddenCommand, AGISyntaxError)
from obelus.common import Handler
from obelus.test import main, watch_logging


def literal_message(text):
    msg = textwrap.dedent(text)
    if not isinstance(msg, bytes):
        # Python 3
        msg = msg.encode('utf-8')
    return msg


HEADER = literal_message("""\
    agi_request: /root/agi_hello.sh
    agi_channel: Local/12345@default-0000000d;2
    agi_language: en
    agi_type: Local
    agi_uniqueid: 1377869716.36
    agi_version: 11.5.0+pf.xivo.13.16~20130722.141054.2668289
    agi_callerid: unknown
    agi_calleridname: unknown
    agi_callingpres: 67
    agi_callingani2: 0
    agi_callington: 0
    agi_callingtns: 0
    agi_dnid: unknown
    agi_rdnis: unknown
    agi_context: default
    agi_extension: 12345
    agi_priority: 1
    agi_enhanced: 0.0
    agi_accountcode:
    agi_threadid: -1257116816
    """)

AGI_ENV = {
    'request': '/root/agi_hello.sh',
    'channel': 'Local/12345@default-0000000d;2',
    'language': 'en',
    'type': 'Local',
    'uniqueid': '1377869716.36',
    'version': '11.5.0+pf.xivo.13.16~20130722.141054.2668289',
    'callerid': 'unknown',
    'calleridname': 'unknown',
    'callingpres': '67',
    'callingani2': '0',
    'callington': '0',
    'callingtns': '0',
    'dnid': 'unknown',
    'rdnis': 'unknown',
    'context': 'default',
    'extension': '12345',
    'priority': '1',
    'enhanced': '0.0',
    'accountcode': '',
    'threadid': '-1257116816',
    }


class AGIProtocolTest(unittest.TestCase):

    def setUp(self):
        self.channel = ProtocolAGIChannel()
        self.proto = AGIProtocol(self.channel)

    def check_header_parsing(self, header, expected_env, expected_args):
        p = self.proto
        for line in header.splitlines(True):
            p.line_received(line)
            self.assertEqual(p._state, 'init')
        p.line_received(b'\n')
        self.assertEqual(p._state, 'idle')
        self.assertEqual(p.env, expected_env)
        self.assertEqual(p.argv, expected_args)

    def proto_idle(self):
        p = self.proto
        p.data_received(HEADER + b"\n")
        self.assertEqual(p._state, 'idle')
        p.channel.write = Mock()
        return p

    def assert_called_once_with_exc(self, mock, exc_class):
        mock.assert_called_once_with(ANY)
        (exc,), _ = mock.call_args
        self.assertIsInstance(exc, exc_class)
        return exc

    @contextlib.contextmanager
    def sending_command(self, args=None):
        p = self.proto
        h = p.send_command(args or ("foo",))
        h.on_result = Mock()
        h.on_exception = Mock()
        yield h

    def test_header_parsing(self):
        self.check_header_parsing(HEADER, AGI_ENV, [])

    def test_header_parsing_with_args(self):
        header = HEADER + literal_message("""\
            agi_arg_1: toto
            agi_arg_2: héhé
            """)
        self.check_header_parsing(header, AGI_ENV, ['toto', 'héhé'])

    def test_parse_result(self):
        f = self.proto._parse_result
        self.assertEqual(f("result=-1"), (-1, {}, None))
        self.assertEqual(f("result=1 endpos=1234"),
                         (1, {'endpos': '1234'}, None))
        self.assertEqual(f("result=0 (foobar)"),
                         (0, {}, "foobar"))
        self.assertEqual(f("result=1 (foobar) endpos=1234"),
                         (1, {'endpos': '1234'}, "foobar"))
        self.assertEqual(f("result=0 (foo quux bar)"),
                         (0, {}, "foo quux bar"))
        self.assertEqual(f("result=1 (foo quux bar) endpos=1234"),
                         (1, {'endpos': '1234'}, "foo quux bar"))

    def test_line_received_when_idle(self):
        p = self.proto_idle()
        # Empty lines are simply ignored, other ones are logged
        p.line_received(b"\n")
        self.assertEqual(p._state, "idle")
        with watch_logging('obelus.agi', level='WARN') as w:
            p.line_received(b"some unexpected data\n")
            self.assertEqual(p._state, "idle")
        self.assertEqual(len(w.output), 1)
        self.assertIn("some unexpected data", w.output[0])

    def test_send_command_invalid(self):
        # Invalid characters in args
        p = self.proto_idle()
        with self.assertRaises(ValueError):
            p.send_command(("foo", "\0"))
        with self.assertRaises(ValueError):
            p.send_command(("foo", "\n"))
        self.assertEqual(0, p.channel.write.call_count)

    def test_send_command_simple_tuple(self):
        p = self.proto_idle()
        h = p.send_command(("set", "variable", "foo", "bar"))
        p.channel.write.assert_called_once_with(b'set variable foo bar\n')
        self.assertIsInstance(h, Handler)

    def test_send_command_simple_list(self):
        p = self.proto_idle()
        h = p.send_command(["set", "variable", "foo", "bar"])
        p.channel.write.assert_called_once_with(b'set variable foo bar\n')
        self.assertIsInstance(h, Handler)

    def test_send_command_escaping_1(self):
        p = self.proto_idle()
        p.send_command(("set", "variable", "some\tspaced data", "bar"))
        p.channel.write.assert_called_once_with(
            b'set variable "some\tspaced data" bar\n')

    def test_send_command_escaping_2(self):
        p = self.proto_idle()
        p.send_command(("set", "variable", 'some"quoted"\\data', "bar"))
        p.channel.write.assert_called_once_with(
            b'set variable "some\\"quoted\\"\\\\data" bar\n')

    def test_send_command_escaping_3(self):
        p = self.proto_idle()
        p.send_command(("say", "alpha", "HELLO WORLD", ""))
        p.channel.write.assert_called_once_with(b'say alpha "HELLO WORLD" ""\n')

    def test_send_command_not_idle(self):
        p = self.proto_idle()
        p.send_command(("foo",))
        with self.assertRaises(RuntimeError):
            p.send_command(("bar",))
        self.assertEqual(p._state, 'awaiting-response')
        p.channel.write.assert_called_once_with(b"foo\n")

    def test_successful_response(self):
        p = self.proto_idle()
        with self.sending_command() as h:
            p.line_received(b"200 result=0 (foobar) endpos=1234\n")
            self.assertEqual(p._state, 'idle')
            h.on_result.assert_called_once_with(
                Response(result=0, variables={'endpos': '1234'}, data='foobar'))
            self.assertEqual(h.on_exception.call_count, 0)

    def test_error_response_200(self):
        p = self.proto_idle()
        with self.sending_command() as h:
            p.line_received(b"200 result=-1\n")
            self.assertEqual(p._state, 'idle')
            self.assertEqual(h.on_result.call_count, 0)
            self.assert_called_once_with_exc(h.on_exception, AGICommandFailure)

    def test_error_response_510(self):
        p = self.proto_idle()
        with self.sending_command() as h:
            p.line_received(b"510 some message\n")
            self.assertEqual(p._state, 'idle')
            self.assertEqual(h.on_result.call_count, 0)
            exc = self.assert_called_once_with_exc(h.on_exception, AGIUnknownCommand)
            self.assertEqual(str(exc), "some message")

    def test_error_response_511(self):
        p = self.proto_idle()
        with self.sending_command() as h:
            p.line_received(b"511 some other message\n")
            self.assertEqual(p._state, 'idle')
            self.assertEqual(h.on_result.call_count, 0)
            exc = self.assert_called_once_with_exc(h.on_exception, AGIForbiddenCommand)
            self.assertEqual(str(exc), "some other message")

    def test_error_response_520(self):
        p = self.proto_idle()
        with self.sending_command() as h:
            p.line_received(b"520 Invalid command syntax.  Proper usage not available.\n")
            self.assertEqual(p._state, 'idle')
            self.assertEqual(h.on_result.call_count, 0)
            exc = self.assert_called_once_with_exc(h.on_exception, AGISyntaxError)
            self.assertEqual(str(exc), "Invalid command syntax.  Proper usage not available.")

    def test_error_response_520_multiline(self):
        p = self.proto_idle()
        with self.sending_command() as h:
            p.line_received(
                b"520-Invalid command syntax.  Proper usage follows:\n")
            p.line_received(
                b"Returns <literal>0</literal> if <replaceable>variablename"
                b"</replaceable> is not set. Returns <literal>1</literal> "
                b"if <replaceable>variablename</replaceable> is set and "
                b"returns the variable in parentheses.\n")
            # The Unicode replacement character (U+FFFD) actually appears
            # in a AGI debug trace on the Asterisk command line. Not sure
            # why, but it's probably worth testing.
            p.line_received(
                b"Example return code: 200 result=1 (testvariable)\xef\xbf\xbd\n")
            self.assertNotEqual(p._state, 'idle')
            p.line_received(
                b"520 End of proper usage.\n")
            self.assertEqual(p._state, 'idle')
            exc = self.assert_called_once_with_exc(h.on_exception, AGISyntaxError)
            self.assertEqual(str(exc), textwrap.dedent("""\
                Invalid command syntax.  Proper usage follows:
                Returns <literal>0</literal> if <replaceable>variablename</replaceable> is not set. Returns <literal>1</literal> if <replaceable>variablename</replaceable> is set and returns the variable in parentheses.
                Example return code: 200 result=1 (testvariable)�
                """))

    def test_multiple_commands(self):
        p = self.proto_idle()
        with self.sending_command(("foo",)) as h:
            p.line_received(b"200 result=0\n")
            h.on_result.assert_called_once_with(
                Response(result=0, variables={}, data=None))
        p.channel.write.assert_called_once_with(b"foo\n")
        with self.sending_command(("bar", "quux")) as h:
            p.line_received(b"200 result=1\n")
            h.on_result.assert_called_once_with(
                Response(result=1, variables={}, data=None))
        p.channel.write.assert_called_with(b"bar quux\n")
        self.assertEqual(p._state, 'idle')




import functools
import logging


class Handler(object):
    """
    A Handler holds callbacks which will be called when an ongoing
    operation terminates.  Also known as Future, Promise, Deferred, etc.

    Generally, you won't create Handler objects yourself.  Instead,
    a producer will create it for you, and you will set the
    :attr:`on_result` and :attr:`on_exception` attributes to be
    notified of the output of the operation.
    """

    _result_cb = None
    _exception_cb = None

    _triggered = False

    @property
    def on_result(self):
        """
        Success callback. Will be called with the handler's result if
        the underlying operation was successful.
        """
        return self._result_cb

    @on_result.setter
    def on_result(self, cb):
        if self._result_cb is not None:
            raise ValueError("on_result already set, cannot override")
        if not callable(cb):
            raise TypeError("on_result should be callable, got %r"
                            % type(cb))
        self._result_cb = cb

    @property
    def on_exception(self):
        """
        Failure callback. Will be called with the handler's exception if
        the underlying operation failed.
        """
        return self._exception_cb

    @on_exception.setter
    def on_exception(self, cb):
        if self._exception_cb is not None:
            raise ValueError("on_exception already set, cannot override")
        if not callable(cb):
            raise TypeError("on_exception should be callable, got %r"
                            % type(cb))
        self._exception_cb = cb

    def set_result(self, result):
        if self._triggered:
            raise RuntimeError("Cannot trigger handler a second time")
        self._triggered = True
        if self._result_cb is not None:
            self._result_cb(result)

    def set_exception(self, exc):
        if self._triggered:
            raise RuntimeError("Cannot trigger handler a second time")
        self._triggered = True
        if self._exception_cb is None:
            raise exc
        else:
            self._exception_cb(exc)

    @classmethod
    def aggregate(cls, handlers):
        """
        Return a new Handler which will trigger when all the given
        handlers have successfully fired, or when one of them fails.
        """
        result_handler = cls()
        n = len(handlers)
        results = [None] * n
        pending = set(range(n))
        def _on_result(i, res):
            if not result_handler._triggered:
                pending.remove(i)
                results[i] = res
                if not pending:
                    result_handler.set_result(results)
        def _on_exception(exc):
            if not result_handler._triggered:
                result_handler.set_exception(exc)
        for i, handler in enumerate(handlers):
            handler.on_result = functools.partial(_on_result, i)
            handler.on_exception = _on_exception
        return result_handler


class LineReceiver(object):
    """
    A base protocol class turning incoming data into distinct lines.
    The `line_received()` method must be defined in order to receive
    individual lines.
    """

    _chunks = None
    logger = logging.getLogger(__name__)

    def data_received(self, data):
        """
        Call this when some *data* (a bytestring) is received.
        """
        if self._chunks is None:
            self._chunks = []
            self._eat_lf = False
        if self._chunks:
            chunks = self._chunks[:-1]
            parts = (self._chunks[-1] + data).splitlines(True)
        else:
            if self._eat_lf and data.startswith(b'\n'):
                data = data[1:]
            self._eat_lf = False
            chunks = []
            parts = data.splitlines(True)
        if len(parts) <= 1 and not data.endswith((b'\r', b'\n')):
            # No new line, just buffer received data
            self._chunks.append(data)
            return
        # First part is the tail of a buffered line
        first_part = parts[0]
        line = b''.join(chunks) + first_part
        last_line = line
        self.line_received(line)
        self._chunks[:] = []
        if len(parts) > 1:
            # Intermediate parts are standalone lines
            for p in parts[1:-1]:
                self.line_received(p)
            # Last part can be an in-progress or complete line
            last_part = parts[-1]
            if last_part.endswith((b'\r', b'\n')):
                last_line = last_part
                self.line_received(last_part)
            else:
                last_line = b''
                self._chunks.append(last_part)
        # Instead of buffering when we receive a line
        # ending with '\r', notify it immediately and swallow the
        # following '\n' later (if any).
        if last_line.endswith(b'\r'):
            self._eat_lf = True



try:
    import twisted
except ImportError:
    twisted = None

if not twisted:
    raise ImportError("Twisted is required for this module to work: "
                      "https://twistedmatrix.com/")

from twisted.internet.protocol import Protocol


class TwistedAdapter(Protocol):
    """
    Twisted adapter for Obelus protocols (e.g. AMIProtocol, AGIProtocol),
    inheriting from twisted.internet.protocol.Protocol.

    Pass a *protocol* instance to create the adapter, which you can
    e.g. return from your Twisted Factory implementation.
    """

    def __init__(self, protocol):
        self.protocol = protocol

    def connectionMade(self):
        self.protocol.connection_made(self)

    def connectionLost(self, failure):
        self.protocol.connection_lost(failure.value)

    def dataReceived(self, data):
        self.protocol.data_received(data)

    # Transport methods
    def write(self, data):
        """
        Write the given *data* bytes on the transport.
        """
        self.transport.write(data)

    def close(self):
        """
        Close the transport's underlying connection.
        """
        self.transport.loseConnection()



"""
Adapter for the Tulip network programming framework.
"""

try:
    import tulip
except ImportError:
    tulip = None

if not tulip:
    raise ImportError("tulip is required for this module to work: "
                      "http://code.google.com/p/tulip/")


if __name__ == "__main__":
    import logging
    import signal

    from .fastagi import FastAGIProtocol, FastAGIExecutor, TCP_PORT
    from .session import AGISession
    from . import examplecli

    parser = examplecli.create_parser(
        description="Tornado-based FastAGI server example")
    parser.add_argument('-p', '--port', type=int, default=TCP_PORT,
                        help='port to listen on')
    parser.add_argument('-L', '--listen', default='127.0.0.1',
                        help='address to listen on')

    options, args = examplecli.parse_args(parser)
    # Tulip's logger is very chatty, dampen it
    logging.getLogger('tulip').setLevel('WARNING')

    class CLIProtocol(examplecli.CLIProtocol, FastAGIProtocol):
        pass

    executor = FastAGIExecutor(CLIProtocol)

    loop = tulip.get_event_loop()
    loop.start_serving(executor.make_protocol, args.listen, args.port)
    loop.add_signal_handler(signal.SIGINT, loop.stop)
    loop.run_forever()




from functools import partial
import logging

from ..common import Handler


class AGISession(object):

    proto = None
    logger = logging.getLogger(__name__)

    def session_established(self):
        """
        Called when the AGI session is established.
        """

    def session_finished(self):
        """
        Called when the AGI session is torn down.
        """

    def run_coroutine(self, gen):
        handler = next(gen)
        self._bind_coroutine_handler(gen, handler)

    def _bind_coroutine_handler(self, gen, handler):
        if not isinstance(handler, Handler):
            raise TypeError(handler.__class__)
        handler.on_result = partial(self._on_coroutine_result, gen, True)
        handler.on_exception = partial(self._on_coroutine_result, gen, False)

    def _on_coroutine_result(self, gen, is_success, result):
        try:
            if is_success:
                handler = gen.send(result)
            else:
                handler = gen.throw(result)
        except StopIteration:
            gen.close()
        else:
            self._bind_coroutine_handler(gen, handler)




import collections
import contextlib
import logging
import sys
from unittest.main import TestProgram
try:
    from unittest import mock
except ImportError:
    import mock
# Allow for e.g. "from mock import Mock"
sys.modules['mock'] = mock


def main():
    """
    Execute the test suite from the current __main__ module.
    """
    TestProgram()


class LoggingWatcher(collections.namedtuple("Watcher", ["records", "output"])):
    pass


class CapturingHandler(logging.Handler):
    """
    A logging handler capturing all (raw and formatted) logging output.
    """

    def __init__(self):
        logging.Handler.__init__(self)
        self.watcher = LoggingWatcher([], [])

    def flush(self):
        pass

    def emit(self, record):
        self.watcher.records.append(record)
        msg = self.format(record)
        self.watcher.output.append(msg)


LOGGING_FORMAT = "%(levelname)s:%(name)s:%(message)s"

@contextlib.contextmanager
def watch_logging(logger_name=None, level=logging.INFO):
    """
    Setup logging to enable log capturing on the given logger
    (or the root logger if None).
    Note that all descendents of this logger will also get captured.
    """
    if isinstance(logger_name, logging.Logger):
        logger = logger_name
    else:
        logger = logging.getLogger(logger_name)
    formatter = logging.Formatter(LOGGING_FORMAT)
    handler = CapturingHandler()
    handler.setFormatter(formatter)
    old_handlers = logger.handlers[:]
    old_level = logger.level
    old_propagate = logger.propagate
    logger.handlers = [handler]
    logger.setLevel(level)
    logger.propagate = False
    try:
        yield handler.watcher
    finally:
        logger.handlers = old_handlers
        logger.propagate = old_propagate
        logger.setLevel(old_level)




from distutils.core import setup

import obelus

version = obelus.__version__

with open('README.rst') as f:
    long_description = f.read()

setup(
    name="obelus",
    version=version,
    packages = ["obelus", "obelus.agi", "obelus.ami", "obelus.test"],
    author="Optiflows",
    author_email="rand@optiflows.com",
    maintainer="Antoine Pitrou",
    maintainer_email="antoine@python.org",
    url="https://pypi.python.org/pypi/obelus/",
    download_url="https://bitbucket.org/optiflowsrd/obelus/get/tip.zip",
    license="MIT",
    description="Protocol implementation of the Asterisk Manager Interface "
                "and Asterisk Gateway Interface",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Topic :: Communications :: Telephony',
        ],
    long_description=long_description,
    keywords=["asterisk", "manager", "gateway", "interface", "AMI", "AGI",
              "twisted", "tornado", "tulip"],
)



import argparse
import logging
import sys

from .protocol import AMIProtocol

log = logging.getLogger(__name__)


class CLIOptions(object):
    pass


class CLIProtocol(AMIProtocol):

    def __init__(self, loop, options):
        AMIProtocol.__init__(self)
        self.loop = loop
        self.options = options

    def connection_made(self, transport):
        log.info("Connection made")
        super(CLIProtocol, self).connection_made(transport)

    def connection_lost(self, exc):
        log.info("Connection lost")
        super(CLIProtocol, self).connection_lost(exc)
        self.loop.stop()

    def greeting_received(self, api_name, api_version):
        log.info("Asterisk greeting: %r, %r", api_name, api_version)
        a = self.send_action('Login', {'username': self.options.username,
                                       'secret': self.options.secret})
        a.on_result = self.login_successful
        a.on_exception = self.login_failed

    def login_successful(self, resp):
        log.info("Successfully logged in")

    def login_failed(self, exc):
        log.error("Failed logging in: %s", exc)
        self.transport.close()


def create_parser(description):
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-H', '--host', default='localhost',
                        help='Asterisk Manager hostname')
    parser.add_argument('-p', '--port', type=int, default=5038,
                        help='Asterisk Manager port number')
    parser.add_argument('-u', '--user', default="xivouser",
                        help='authentication username')
    parser.add_argument('-s', '--secret', default="xivouser",
                        help='authentication secret')

    parser.add_argument('-q', '--quiet', action='store_true',
                        help='hide information messages')
    parser.add_argument('--debug', action='store_true',
                        help='show debug messages')

    return parser


def parse_args(parser):
    args = parser.parse_args()
    options = CLIOptions()
    options.host, options.port = args.host, args.port
    options.username, options.secret = args.user, args.secret

    if args.quiet:
        level = logging.WARNING
    elif args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(stream=sys.stderr, level=level)

    return options, args
